#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import math
from typing import Any, List, Optional, Tuple

import torch
from torch import Tensor

from natten.types import CausalArgType, DimensionType


class Problem:
    def __init__(
        self,
        na_dim: int,
        batch_size: int,
        heads: int,
        dim: int,
        input_size: DimensionType,
        window_size: DimensionType,
        stride: DimensionType,
        dilation: DimensionType,
        dtype: torch.dtype,
        is_causal: CausalArgType,
        additional_kv_length: Optional[int] = None,
        dim_value: Optional[int] = None,
    ):
        self.na_dim = na_dim
        self.batch_size = batch_size
        self.heads = heads
        self.dim = dim
        self.dim_value = dim_value or dim
        assert (
            len(window_size)
            == len(stride)
            == len(dilation)
            == len(input_size)
            == self.na_dim
        )
        self.input_size = input_size
        self.window_size = window_size
        self.stride = stride
        self.dilation = dilation
        self.dtype = dtype
        self.is_causal = is_causal or [False for _ in range(na_dim)]
        self.has_additional_kv = (
            additional_kv_length is not None and additional_kv_length > 0
        )
        self.additional_kv_length = additional_kv_length

    @property
    def is_self_attn(self) -> bool:
        return not any(c for c in self.is_causal) and all(
            x == w for x, w in zip(self.input_size, self.window_size)
        )

    def get_q_tensor_shape(self, heads_last: bool, flatten: bool = False) -> List:
        token_layout_shape = (
            [x for x in self.input_size]
            if not flatten
            else [math.prod(self.input_size)]
        )
        if not heads_last:
            return [self.batch_size, self.heads] + token_layout_shape + [self.dim]
        return [self.batch_size] + token_layout_shape + [self.heads, self.dim]

    def get_k_tensor_shape(self, heads_last: bool, flatten: bool = False) -> List:
        return self.get_q_tensor_shape(heads_last=heads_last, flatten=flatten)

    def get_v_tensor_shape(self, heads_last: bool, flatten: bool = False) -> List:
        token_layout_shape = (
            [x for x in self.input_size]
            if not flatten
            else [math.prod(self.input_size)]
        )
        if not heads_last:
            return [self.batch_size, self.heads] + token_layout_shape + [self.dim_value]
        return [self.batch_size] + token_layout_shape + [self.heads, self.dim_value]

    def get_o_tensor_shape(self, heads_last: bool, flatten: bool = False) -> List:
        return self.get_v_tensor_shape(heads_last=heads_last, flatten=flatten)

    def get_add_k_tensor_shape(self, heads_last: bool) -> List:
        assert self.additional_kv_length is not None
        if not heads_last:
            return [self.batch_size, self.heads, self.additional_kv_length, self.dim]
        return [self.batch_size, self.additional_kv_length, self.heads, self.dim]

    def get_add_v_tensor_shape(self, heads_last: bool) -> List:
        assert self.additional_kv_length is not None
        if not heads_last:
            return [
                self.batch_size,
                self.heads,
                self.additional_kv_length,
                self.dim_value,
            ]
        return [self.batch_size, self.additional_kv_length, self.heads, self.dim_value]

    def make_qkv_tensors(
        self, device, requires_grad: bool, heads_last: bool, flatten: bool = False
    ) -> Tuple[Tensor, Tensor, Tensor]:
        q = torch.randn(
            self.get_q_tensor_shape(heads_last=heads_last, flatten=flatten),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )
        k = torch.randn(
            self.get_k_tensor_shape(heads_last=heads_last, flatten=flatten),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )
        v = torch.randn(
            self.get_v_tensor_shape(heads_last=heads_last, flatten=flatten),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )

        return q, k, v

    def make_qkvo_tensors(
        self, device, requires_grad: bool, heads_last: bool, flatten: bool = False
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        q, k, v = self.make_qkv_tensors(
            device=device,
            requires_grad=requires_grad,
            heads_last=heads_last,
            flatten=flatten,
        )
        o = torch.randn(
            self.get_o_tensor_shape(heads_last=heads_last, flatten=flatten),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )
        return q, k, v, o

    def make_additional_kv_tensors(
        self, device, requires_grad: bool, heads_last: bool
    ) -> Tuple[Tensor, Tensor]:
        add_k = torch.randn(
            self.get_add_k_tensor_shape(heads_last=heads_last),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )
        add_v = torch.randn(
            self.get_add_v_tensor_shape(heads_last=heads_last),
            device=device,
            dtype=self.dtype,
            requires_grad=requires_grad,
        )

        return add_k, add_v

    def __str__(self):
        return (
            "Problem("
            + f"batch_size={self.batch_size}, "
            + f"heads={self.heads}, "
            + f"input_size={self.input_size}, "
            + f"window_size={self.window_size}, "
            + f"stride={self.stride}, "
            + f"dilation={self.dilation}, "
            + f"is_causal={self.is_causal}, "
            + f"additional_kv_length={self.additional_kv_length}, "
            + f"dtype={self.dtype})"
        )

    def __repr__(self):
        return self.__str__()


def generate_problem(
    batch_size: int,
    heads: int,
    input_size: DimensionType,
    dim: int,
    window_size: DimensionType,
    stride: DimensionType,
    dilation: DimensionType,
    dtype: Any,
    is_causal: CausalArgType,
    additional_kv_length: Optional[int] = None,
    dim_value: Optional[int] = None,
) -> Problem:
    na_dim = len(input_size)
    assert len(window_size) == na_dim
    assert len(stride) == na_dim
    assert len(dilation) == na_dim
    return Problem(
        na_dim=na_dim,
        batch_size=batch_size,
        heads=heads,
        dim=dim,
        dim_value=dim_value,
        input_size=input_size,
        window_size=window_size,
        stride=stride,
        dilation=dilation,
        dtype=dtype,
        is_causal=is_causal,
        additional_kv_length=additional_kv_length,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from enum import Enum
from typing import Dict, List, Optional, Tuple

import torch
from torch.profiler import profile as torch_profile


class LibNattenOp(Enum):
    Unrecognized = 0

    # Fused
    FnaForward = 1
    FnaBackward = 2

    # FMHA
    FmhaForward = 3
    FmhaBackward = 4

    # Misc
    Reduction = 5
    Elementwise = 6


NATTEN_TAGS = {
    LibNattenOp.FnaForward: [
        "natten::cuda::fna::FusedNeighborhoodAttentionKernel",
        "cutlass::fna::kernel::Sm100FnaFwdKernelTmaWarpspecialized",
        "cutlass::fna::collective::FnaMainloopTmaWarpSpecializedSm90",
        "cutlass::fna::collective::FnaMainloopTmaSm90",
    ],
    LibNattenOp.FnaBackward: [
        "natten::cuda::fna::FusedNeighborhoodAttentionBackwardKernel<",
        "cutlass::fna::collective::FnaBwdMainloopTmaWarpSpecializedSm90",
        "cutlass::fna::kernel::Sm100FnaBwdKernelTmaWarpSpecialized",
    ],
    LibNattenOp.FmhaForward: [
        "cutlass::fmha::kernel::Sm100FmhaFwdKernelTmaWarpspecialized",
        "cutlass::fmha::collective::FmhaMainloopTmaWarpSpecializedSm90",
        "cutlass::fmha::collective::FmhaMainloopTmaSm90",
        "natten::cuda::fmha::FusedNeighborhoodAttentionKernel",
        "natten::cuda::fmha::AttentionKernel",
    ],
    LibNattenOp.FmhaBackward: [
        "cutlass::fmha::kernel::Sm100FmhaFwdKernelTmaWarpspecialized",
        "cutlass::fmha::collective::FmhaMainloopTmaWarpSpecializedSm90",
        "cutlass::fmha::collective::FmhaMainloopTmaSm90",
        "natten::cuda::fmha::AttentionBackwardKernel",
        "cutlass::fmha::collective::FmhaBwdMainloopTmaWarpSpecializedSm90",
        "cutlass::fmha::kernel::Sm100FmhaBwdKernelTmaWarpSpecialized",
    ],
    LibNattenOp.Reduction: [
        "natten::cuda::reduction::kernel::ComputeDelta",
        "cutlass::fmha::kernel::FmhaKernelBwdSumOdO",
    ],
    LibNattenOp.Elementwise: [
        "cutlass::fmha::kernel::FmhaKernelBwdConvert",
    ],
}


def _format_time(time_us: float) -> str:
    """
    Source: https://github.com/pytorch/pytorch/blob/orig/release/1.13/torch/autograd/profiler_util.py
    """
    US_IN_SECOND = 1000.0 * 1000.0
    US_IN_MS = 1000.0
    if time_us >= US_IN_SECOND:
        return "{:.3f}s".format(time_us / US_IN_SECOND)
    if time_us >= US_IN_MS:
        return "{:.3f}ms".format(time_us / US_IN_MS)
    return "{:.3f}us".format(time_us)


class Result:
    def __init__(
        self,
        op: Optional[LibNattenOp],
        op_str: str,
        time: float,
        index: int = -1,
        num_calls: int = 1,
        tag: str = "N/A",
        framework: str = "N/A",
        kernel_type: str = "N/A",
    ):
        self.framework = framework
        self.kernel_type = kernel_type
        self.tag = "N/A" if not tag else tag

        if op in [
            LibNattenOp.FnaForward,
            LibNattenOp.FnaBackward,
            LibNattenOp.FmhaForward,
            LibNattenOp.FmhaBackward,
        ]:
            self.kernel_type = "attention"
            self.framework = "CUTLASS"

        elif op in [LibNattenOp.Reduction]:
            self.kernel_type = "reduction"
            self.framework = "CUTLASS"
            self.tag = "-"

        elif op in [LibNattenOp.Elementwise]:
            self.kernel_type = "elementwise"
            self.framework = "CUTLASS"
            self.tag = "-"

        self.op = op
        if op is not None and isinstance(op, LibNattenOp):
            self.op_str = op.name
        else:
            self.op_str = op_str

        self.time = time
        self.index = index
        self.num_calls = num_calls

    def to_str_name(self):
        return (
            f"framework_{self.framework}_type{self.kernel_type}_"
            + f"tag_{self.tag}_op_str{self.op_str}"
        )

    @property
    def op_name_formatted(self):
        op_str = "-" if self.op_str is None else str(self.op_str)
        MAX_CHARS = 45
        if len(op_str) > MAX_CHARS:
            return op_str[:MAX_CHARS] + "..."
        return op_str

    @property
    def time_str(self) -> str:
        return _format_time(self.time)

    def __le__(self, other) -> bool:
        return self.index <= other.index

    def __ge__(self, other) -> bool:
        return self.index >= other.index

    def __lt__(self, other) -> bool:
        return self.index < other.index

    def __gt__(self, other) -> bool:
        return self.index > other.index

    def __eq__(self, other) -> bool:
        return self.index == other.index

    def __ne__(self, other) -> bool:
        return self.index != other.index

    def __radd__(self, other):
        return self

    def __add__(self, other):
        return Result(
            op=self.op if self.op == other.op else "INV",
            op_str=self.op_str if self.op_str == other.op_str else "INV",
            time=self.time + other.time,
            index=min(self.index, other.index),
            num_calls=self.num_calls + other.num_calls,
            tag=self.tag if self.tag == other.tag else "INV",
            framework=self.framework if self.framework == other.framework else "INV",
            kernel_type=(
                self.kernel_type if self.kernel_type == other.kernel_type else "INV"
            ),
        )


def get_natten_op(sym: str) -> Optional[LibNattenOp]:
    for natten_op_type, tags in NATTEN_TAGS.items():
        for tag in tags:
            if tag in sym:
                return natten_op_type
    return None


KERNEL_TYPE_LOOKUP = {
    "attention": ["sdpa", "fmha", "flash_attn"],
    "gemm": ["gemm", "matmul"],
    "gemv": ["gemv"],
    "elementwise": ["elementwise"],
    "softmax": ["softmax"],
}


# TODO: I really hate how we're parsing symbols in this entire file!
def strip_str_name(name: str) -> Tuple[str, str]:
    if name.startswith("cudnn_generated_fort_native_"):
        return "attention", name.replace("cudnn_generated_fort_native_sdpa_", "")

    def remove_wrapper(s, wrapper):
        if s.startswith(wrapper):
            return s[len(wrapper) :], True
        return s, False

    kernel_type = "-"
    found = False
    for kernel_type_, tag_list in KERNEL_TYPE_LOOKUP.items():
        for tag in tag_list:
            if tag in name.lower():
                kernel_type = kernel_type_
                found = True
                break
        if found:
            break

    name = name.strip()
    # name, is_cutlass = remove_wrapper(name, "void cutlass::Kernel<")
    name, _ = remove_wrapper(name, "void (anonymous namespace)::")
    name, _ = remove_wrapper(name, "void ")
    name = name.split("<")[0].split(">")[0]
    name = name.replace("(anonymous namespace)::", "")
    name = name.replace("(anonymous namespace)", "")
    name = name.replace("at::native::", "")

    if name.startswith("cudnn_generated_fort_native_"):
        name = name.replace("cudnn_generated_fort_native_", "")
    if name.startswith("triton_"):
        name = name.replace("triton_", "")

    namespace_split = name.split("::")

    if len(namespace_split) <= 1:
        return kernel_type, name

    return kernel_type, ".".join(namespace_split)


ARCH_LOOKUP = {
    "Sm50": ["sm50", "maxwell"],
    "Sm52": ["sm52"],
    "Sm53": ["sm53"],
    "Sm60": ["sm60", "pascal"],
    "Sm61": ["sm61"],
    "Sm62": ["sm62"],
    "Sm70": ["sm70", "volta"],
    "Sm72": ["sm72"],
    "Sm75": ["sm75", "turing"],
    "Sm80": ["sm80"],
    "Sm86": ["sm86"],
    "Sm87": ["sm87"],
    "Sm8X": ["ampere"],
    "Sm89": ["sm89"],
    "Sm90": ["sm90", "hopper"],
    "Sm100": ["sm100", "blackwell"],
    "Sm101": ["sm101"],
    "Sm120": ["sm120"],
}


FRAMEWORK_LOOKUP = {
    "CUTLASS": ["cutlass", "cute"],
    "cuDNN": ["cudnn"],
    "cuBLAS": ["cublas"],
    "PyTorch": ["c10", "aten", "at::native"],
    "Triton": ["triton"],
}


def get_arch(sym: str) -> str:
    for arch, tag_list in ARCH_LOOKUP.items():
        for tag in tag_list:
            if tag in sym.lower():
                return arch
    return "-"


def get_framework(sym: str) -> str:
    for k, tag_list in FRAMEWORK_LOOKUP.items():
        for tag in tag_list:
            if tag in sym.lower():
                return k
    return "-"


def get_op_prio(op: Optional[LibNattenOp]) -> int:
    if op is None or not isinstance(op, LibNattenOp):
        return 999

    return int(op.value)


def convert_to_natten_profiler_ops(
    profiler: torch_profile,
) -> Optional[List[Result]]:

    events = profiler.events()

    results: Dict[str, Result] = {}
    for evt in events:
        if evt.key.startswith("Memcpy") or evt.key.startswith("Memset"):
            continue

        time_total = (
            evt.device_time_total if torch.cuda.is_available() else evt.cpu_time_total
        )

        natten_op = get_natten_op(evt.key)
        arch_tag = get_arch(evt.key)
        framework = get_framework(evt.key)
        kernel_type, op_str = strip_str_name(evt.key)

        if time_total > 0:
            result = Result(
                op=natten_op,
                op_str=op_str,
                time=time_total,
                index=get_op_prio(natten_op),  # Put NATTEN ops at the top
                num_calls=1,
                tag=arch_tag,
                framework=framework,
                kernel_type=kernel_type,
            )

            key = result.to_str_name()
            if key in results:
                results[key] += result
            else:
                results[key] = result

    results_final = [v for v in results.values()]
    return None if results_final is None else sorted(results_final)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import enum
import logging
import os

log_format = "| %(asctime)s | [[ %(name)s ]] [ %(levelname)s ]: %(message)s"


class LogLevel(enum.Enum):
    Default = 0
    Debug = 1
    Info = 2
    Warnings = 3
    Errors = 4
    Critical = 5


def _get_log_level() -> LogLevel:
    if "NATTEN_LOG_LEVEL" in os.environ:
        log_level = str(os.environ["NATTEN_LOG_LEVEL"]).lower()
        if log_level == "debug":
            return LogLevel.Debug
        elif log_level == "info":
            return LogLevel.Info
        elif log_level == "warning":
            return LogLevel.Warnings
        elif log_level == "error":
            return LogLevel.Errors
        elif log_level == "critical":
            return LogLevel.Critical

    return LogLevel.Default


_map_log_level = {
    LogLevel.Default: logging.INFO,
    LogLevel.Debug: logging.DEBUG,
    LogLevel.Info: logging.INFO,
    LogLevel.Warnings: logging.WARNING,
    LogLevel.Errors: logging.ERROR,
    LogLevel.Critical: logging.CRITICAL,
}


def get_logger(name):
    logger = logging.getLogger(name)
    log_level = _map_log_level[_get_log_level()]
    logger.setLevel(log_level)
    formatter = logging.Formatter(log_format)
    handler = logging.StreamHandler()
    handler.setLevel(log_level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import itertools
import math
from typing import List, Optional, Tuple

import torch
from torch import Tensor

from ....context import (
    is_kv_parallelism_in_fused_na_enabled,
    is_memory_usage_strict,
    is_memory_usage_unrestricted,
)
from ....types import (
    CutlassFmhaBackwardConfigType,
    CutlassFmhaForwardConfigType,
    CutlassFnaBackwardConfigType,
    CutlassFnaForwardConfigType,
    DimensionType,
)
from ....utils.checks import check_dilation_arg, check_tile_shape
from ....utils.device import get_device_cc, is_cuda

from ....utils.tuples import ceil_div_int, ceil_div_tuple

# FNA/FMHA forward supports 64x64 and 32x128 GEMM configs in all
# use cases. Some architectures (SM80 and SM90 )have more shared
# memory so they can handle 64x128 GEMMs.

from .fna_backward_128x128 import _FNA_BACKWARD_128x128_TILE_SIZES
from .fna_backward_128x64 import _FNA_BACKWARD_128x64_TILE_SIZES
from .fna_backward_64x64 import _FNA_BACKWARD_64x64_TILE_SIZES

# FNA/FMHA backward supports 64x64 GEMM configs in all
# use cases. Some architectures have more shared memory
# so they can handle 128x64 or 128x128 GEMMs, but that
# is also dependent on the GEMM K.

from .fna_forward_32x128 import _FNA_FORWARD_32x128_TILE_SIZES
from .fna_forward_64x128 import _FNA_FORWARD_64x128_TILE_SIZES
from .fna_forward_64x64 import _FNA_FORWARD_64x64_TILE_SIZES


def _get_default_tile_shapes_forward(
    na_dim: int,
) -> CutlassFnaForwardConfigType:
    assert na_dim in [1, 2, 3]

    if na_dim == 1:
        return ((64,), (64,))
    if na_dim == 2:
        return ((8, 8), (8, 8))
    if na_dim == 3:
        return ((4, 4, 4), (4, 4, 4))

    raise NotImplementedError()


def get_all_tile_shapes_forward(
    input_tensor: Tensor,
) -> List[CutlassFnaForwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    device = input_tensor.device

    if not is_cuda(device):
        return []

    # SM80 and SM90 have more shared memory than SM86 and SM89
    # and their tensor core GEMMs can therefore target larger
    # tile shapes.
    # SM80 and 86 have been tested, but I don't have an SM89.
    # However, I suspect SM89 is to SM90 what SM86 was to SM80
    # in terms of shared memory (and only that).
    # Better to disable the larger tile configs for SM89 as well
    # as 86 until we can test it.
    if get_device_cc(device) in [86, 89]:
        return (
            _FNA_FORWARD_32x128_TILE_SIZES[na_dim]
            + _FNA_FORWARD_64x64_TILE_SIZES[na_dim]
        )

    return (
        _FNA_FORWARD_32x128_TILE_SIZES[na_dim]
        + _FNA_FORWARD_64x64_TILE_SIZES[na_dim]
        + _FNA_FORWARD_64x128_TILE_SIZES[na_dim]
    )


# For FMHA
def get_all_tile_sizes_forward(
    input_tensor: Tensor,
) -> List[CutlassFmhaForwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    tile_shapes = get_all_tile_shapes_forward(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in tile_shapes)

    tile_sizes = [(q_t[0], kv_t[0]) for q_t, kv_t in tile_shapes]

    return tile_sizes


def get_default_forward_config(
    input_tensor: Tensor, dilation: Optional[DimensionType] = None
) -> CutlassFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    return _get_default_tile_shapes_forward(na_dim)


get_all_forward_configs = get_all_tile_shapes_forward
get_all_fmha_forward_configs = get_all_tile_sizes_forward


def check_cutlass_fna_forward_config(
    input_tensor: Tensor,
    dilation: Optional[DimensionType] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
) -> CutlassFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    if q_tile_shape is None and kv_tile_shape is None:
        return get_default_forward_config(input_tensor=input_tensor, dilation=dilation)

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    tile_shapes = get_all_tile_shapes_forward(input_tensor=input_tensor)

    for q_t, kv_t in tile_shapes:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            return (q_t, kv_t)  # type: ignore

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_shapes):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS FNA-{na_dim}D. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(tile_shapes)} configurations implementable "
        f"with CUTLASS 2.X FNA for SM{device_cc}. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_cutlass_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_fmha_forward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> CutlassFmhaForwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    if q_tile_size is None and kv_tile_size is None:
        q_tile_shape, kv_tile_shape = get_default_forward_config(
            input_tensor=input_tensor
        )
        assert len(q_tile_shape) == len(kv_tile_shape) == 1
        return (q_tile_shape[0], kv_tile_shape[0])

    tile_sizes = get_all_tile_sizes_forward(input_tensor=input_tensor)

    for q_t, kv_t in tile_sizes:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            return (q_t, kv_t)

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_sizes):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS FMHA. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(tile_sizes)} configurations implementable "
        f"with CUTLASS 2.X FMHA for SM{device_cc}. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_cutlass_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


###### Backward


def _get_default_tile_shapes_backward(
    na_dim: int,
) -> Tuple[DimensionType, DimensionType]:
    assert na_dim in [1, 2, 3]

    if na_dim == 1:
        return ((64,), (64,))
    if na_dim == 2:
        return ((8, 8), (8, 8))
    if na_dim == 3:
        return ((4, 4, 4), (4, 4, 4))

    raise NotImplementedError()


def get_all_tile_shapes_backward(
    input_tensor: Tensor,
) -> List[CutlassFnaForwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    device = input_tensor.device
    dtype = input_tensor.dtype
    dim_per_head = input_tensor.shape[-1]

    if not is_cuda(device):
        return []

    compute_cap = get_device_cc(device)

    assert dtype in [torch.float32, torch.float16, torch.bfloat16]

    if dtype == torch.float32 and compute_cap not in [80, 90]:
        return _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]

    elif dtype == torch.float32:
        return (
            _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]
            + _FNA_BACKWARD_128x64_TILE_SIZES[na_dim]
        )

    if compute_cap == 70:
        return (
            _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]
            + _FNA_BACKWARD_128x64_TILE_SIZES[na_dim]
        )

    if compute_cap in [80, 90] and dim_per_head <= 128:
        return (
            _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]
            + _FNA_BACKWARD_128x128_TILE_SIZES[na_dim]
        )
    elif compute_cap in [80, 90]:
        return (
            _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]
            + _FNA_BACKWARD_128x64_TILE_SIZES[na_dim]
        )

    return _FNA_BACKWARD_64x64_TILE_SIZES[na_dim]


# For FMHA
def get_all_tile_sizes_backward(
    input_tensor: Tensor,
) -> List[CutlassFmhaForwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    tile_shapes = get_all_tile_shapes_backward(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in tile_shapes)

    tile_sizes = [(q_t[0], kv_t[0]) for q_t, kv_t in tile_shapes]

    return tile_sizes


def _get_max_grid_size_allowed() -> int:
    if is_memory_usage_unrestricted():
        return 65535
    if is_memory_usage_strict():
        return 1024

    return 4096


def get_min_splits(na_dim: int) -> DimensionType:
    assert na_dim in [1, 2, 3]
    return tuple(1 for _ in range(na_dim))  # type: ignore


def get_max_splits(
    input_shape: DimensionType, dilation: DimensionType, kv_tile_shape: DimensionType
) -> DimensionType:
    extent_per_dilation_group = ceil_div_tuple(input_shape, dilation)
    return tuple(
        ceil_div_int(x, t) for x, t in zip(extent_per_dilation_group, kv_tile_shape)
    )  # type: ignore


def _reduce_max_kv_splits(
    na_dim: int,
    kv_splits: DimensionType,
    max_splits: int,
) -> DimensionType:
    assert isinstance(kv_splits, tuple)
    assert na_dim in [1, 2, 3]

    if na_dim == 1:
        assert len(kv_splits) == 1
        return (min(kv_splits[0], max_splits),)

    if na_dim == 2:
        assert len(kv_splits) == 2
        splits_x = max(min(max_splits // 2, kv_splits[0]), 1)
        splits_y = max(min(max_splits // splits_x, kv_splits[1]), 1)
        assert (
            0 < splits_x * splits_y <= max_splits
        ), f"{splits_x=} * {splits_y=} does not fall in range [0, {max_splits}]"
        return (splits_x, splits_y)

    if na_dim == 3:
        assert len(kv_splits) == 3
        splits_x = max(min(max_splits // 3, kv_splits[0]), 1)
        splits_y = max(min(max_splits // splits_x, kv_splits[1]), 1)
        splits_z = max(min(max_splits // (splits_x * splits_y), kv_splits[2]), 1)
        assert (
            0 < splits_x * splits_y * splits_z <= max_splits
        ), f"{splits_x=} * {splits_y=} * {splits_z=} does not fall in range [0, {max_splits}]"
        return (splits_x, splits_y, splits_z)

    raise NotImplementedError()


def _get_possible_kv_splits(
    min_splits: DimensionType,
    max_splits: DimensionType,
):
    assert 0 < len(min_splits) == len(max_splits) < 4
    na_dim = len(max_splits)
    if na_dim == 1:
        return itertools.product(
            range(min_splits[0], max_splits[0] + 1),
        )
    if na_dim == 2:
        assert len(min_splits) == len(max_splits) == 2
        return itertools.product(
            range(min_splits[0], max_splits[0] + 1),
            range(min_splits[1], max_splits[1] + 1),
        )
    if na_dim == 3:
        assert len(min_splits) == len(max_splits) == 3
        return itertools.product(
            range(min_splits[0], max_splits[0] + 1),
            range(min_splits[1], max_splits[1] + 1),
            range(min_splits[2], max_splits[2] + 1),
        )

    raise NotImplementedError()


def get_default_kv_splits_backward(
    input_tensor: Tensor,
    kv_tile_shape: DimensionType,
    dilation: Optional[DimensionType] = None,
) -> DimensionType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    dilation = check_dilation_arg(na_dim, dilation)
    input_shape: DimensionType = tuple(int(x) for x in input_tensor.shape[1 : na_dim + 1])  # type: ignore

    assert na_dim in [1, 2, 3]
    if na_dim == 1:
        kv_splits: DimensionType = (1,)
    elif na_dim == 2:
        kv_splits = (1, 1)

    elif na_dim == 3:
        kv_splits = (1, 1, 1)

    if (
        is_kv_parallelism_in_fused_na_enabled()
        and not torch.are_deterministic_algorithms_enabled()
    ):
        kv_splits = get_max_splits(
            input_shape, dilation=dilation, kv_tile_shape=kv_tile_shape
        )
        total_kv_splits = math.prod(kv_splits)

        batch_size = input_tensor.shape[0]
        num_heads = input_tensor.shape[-2]
        num_dilation_splits = math.prod(dilation)
        max_kv_splits_allowed = max(
            1,
            _get_max_grid_size_allowed()
            // (batch_size * num_heads * num_dilation_splits),
        )

        if total_kv_splits > max_kv_splits_allowed:
            kv_splits = _reduce_max_kv_splits(
                na_dim=na_dim, kv_splits=kv_splits, max_splits=max_kv_splits_allowed
            )

    return kv_splits


def get_default_backward_config(
    input_tensor: Tensor, dilation: Optional[DimensionType] = None
) -> CutlassFnaBackwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    dilation = check_dilation_arg(na_dim, dilation)

    q_tile_shape, kv_tile_shape = _get_default_tile_shapes_backward(na_dim)
    kv_splits = get_default_kv_splits_backward(
        input_tensor=input_tensor, kv_tile_shape=kv_tile_shape, dilation=dilation
    )
    use_pt_reduction = False
    return (q_tile_shape, kv_tile_shape, kv_splits, use_pt_reduction)  # type: ignore


# WARNING: use with caution
# There's quite a lot of configurations possible, especially if we don't restrict
# KV parallelism.
def get_all_backward_configs(
    input_tensor: Tensor, dilation: Optional[DimensionType] = None
) -> List[CutlassFnaBackwardConfigType]:

    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    dilation = check_dilation_arg(na_dim, dilation)

    input_shape: DimensionType = tuple(x for x in input_tensor.shape[1 : na_dim + 1])  # type: ignore
    assert len(input_shape) == na_dim and all(isinstance(x, int) for x in input_shape)
    batch_size = input_tensor.shape[0]
    num_heads = input_tensor.shape[-2]

    possible_tile_shapes = get_all_tile_shapes_backward(input_tensor)

    possible_configs = []
    assert len(input_shape) == len(dilation) == na_dim
    num_dilation_splits = math.prod(dilation)
    max_kv_splits_allowed = (
        1
        if not is_kv_parallelism_in_fused_na_enabled()
        else max(
            1,
            _get_max_grid_size_allowed()
            // (batch_size * num_heads * num_dilation_splits),
        )
    )
    for query_tile_shape, kv_tile_shape in possible_tile_shapes:
        min_kv_splits = get_min_splits(na_dim)
        max_kv_splits = get_max_splits(
            input_shape, dilation=dilation, kv_tile_shape=kv_tile_shape
        )
        max_kv_splits_total = math.prod(max_kv_splits)
        if max_kv_splits_total > max_kv_splits_allowed:
            max_kv_splits = _reduce_max_kv_splits(
                na_dim=na_dim, kv_splits=max_kv_splits, max_splits=max_kv_splits_allowed
            )

        # Potential duplicates
        if math.prod(max_kv_splits) > 1:
            for kv_splits in _get_possible_kv_splits(min_kv_splits, max_kv_splits):
                for use_pt_reduction in [False, True]:
                    possible_configs.append(
                        (
                            query_tile_shape,
                            kv_tile_shape,
                            kv_splits,
                            use_pt_reduction,
                        )
                    )
        else:
            # min_kv_splits == max_kv_splits
            for use_pt_reduction in [False, True]:
                possible_configs.append(
                    (query_tile_shape, kv_tile_shape, min_kv_splits, use_pt_reduction)
                )

    return possible_configs  # type: ignore


def get_all_fmha_backward_configs(
    input_tensor: Tensor, dilation: Optional[DimensionType] = None
) -> List[CutlassFmhaBackwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    if dilation is not None:
        raise ValueError(f"FMHA does not support dilation, got {dilation=}.")

    backward_configs = get_all_backward_configs(input_tensor)
    assert all(
        len(q_t) == len(kv_t) == len(kv_s) == 1
        for q_t, kv_t, kv_s, _ in backward_configs
    )

    fmha_backward_configs = [
        (q_t[0], kv_t[0], kv_s[0], use_pt_red)
        for q_t, kv_t, kv_s, use_pt_red in backward_configs
    ]

    return fmha_backward_configs


def check_cutlass_fna_backward_config(
    input_tensor: Tensor,
    dilation: Optional[DimensionType] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    kv_splits: Optional[DimensionType] = None,
    use_pt_reduction: bool = False,
) -> CutlassFnaBackwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim
    dilation = check_dilation_arg(na_dim, dilation)

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    if q_tile_shape is None and kv_tile_shape is None:
        q_tile_shape, kv_tile_shape, _, _ = get_default_backward_config(
            input_tensor=input_tensor, dilation=dilation
        )

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    tile_shapes = get_all_tile_shapes_backward(input_tensor)

    for q_t, kv_t in tile_shapes:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            kv_splits = kv_splits or get_default_kv_splits_backward(
                input_tensor=input_tensor,
                kv_tile_shape=kv_tile_shape,
                dilation=dilation,
            )
            return (q_t, kv_t, kv_splits, use_pt_reduction)  # type: ignore

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_shapes):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS FNA-{na_dim}D. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(tile_shapes)} configurations implementable "
        f"with CUTLASS 2.X FNA for SM{device_cc}, with input tensor shape "
        f"{input_tensor.shape} and {dilation=}. Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_cutlass_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_fmha_kv_splits(
    kv_splits: Optional[int], input_tensor: Tensor, kv_tile_size: int
) -> int:
    if kv_splits is not None and isinstance(kv_splits, int):
        return kv_splits

    if kv_splits is None:
        default_kv_splits: DimensionType = get_default_kv_splits_backward(
            input_tensor=input_tensor, kv_tile_shape=(kv_tile_size,)
        )
        assert len(default_kv_splits) == 1
        return default_kv_splits[0]

    raise ValueError(f"Invalid type {type(kv_splits)} for kv_splits.")


def check_cutlass_fmha_backward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    kv_splits: Optional[int] = None,
    use_pt_reduction: bool = False,
) -> CutlassFmhaBackwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    if q_tile_size is None and kv_tile_size is None:
        q_tile_shape, kv_tile_shape, _, _ = get_default_backward_config(
            input_tensor=input_tensor
        )
        assert len(q_tile_shape) == len(kv_tile_shape) == 1
        q_tile_size, kv_tile_size = q_tile_shape[0], kv_tile_shape[0]

    tile_sizes = get_all_tile_sizes_backward(input_tensor)

    for q_t, kv_t in tile_sizes:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            kvs = check_fmha_kv_splits(
                kv_splits, input_tensor=input_tensor, kv_tile_size=kv_t
            )
            return (q_t, kv_t, kvs, use_pt_reduction)

    # Fail and make suggestions
    device_cc = get_device_cc(input_tensor.device)
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(tile_sizes):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS FMHA. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(tile_sizes)} configurations implementable "
        f"with CUTLASS 2.X FNA for SM{device_cc}, with input tensor shape "
        f"{input_tensor.shape}. Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_cutlass_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
from enum import Enum

import torch

from .utils import log


logger = log.get_logger(__name__)


class MemoryUsagePreference(Enum):
    Default = 0
    Strict = 1
    Unrestricted = 2


class NattenContext:
    is_deterministic_mode_enabled: bool = False
    is_kv_parallelism_enabled: bool = True
    training_memory_preference: MemoryUsagePreference = MemoryUsagePreference.Default
    flex_compile_allowed: bool = False
    flex_compile_backprop_allowed: bool = False

    @staticmethod
    def reset():
        NattenContext.is_deterministic_mode_enabled = False
        NattenContext.is_kv_parallelism_enabled = True
        NattenContext.training_memory_preference = MemoryUsagePreference.Default
        NattenContext.flex_compile_allowed = False
        NattenContext.flex_compile_backprop_allowed = False


def set_memory_usage_preference(pref: str = "default"):
    """Sets memory usage preference for KV parallelism in `"cutlass-fna"` and `"cutlass-fmha"`
    backends.

    Args:
        pref: Choices are `"default"`, `"strict"`, and `"unrestricted"`.
    """
    if pref == "default":
        NattenContext.training_memory_preference = MemoryUsagePreference.Default
    elif pref == "strict":
        NattenContext.training_memory_preference = MemoryUsagePreference.Strict
    elif pref == "unrestricted":
        NattenContext.training_memory_preference = MemoryUsagePreference.Unrestricted
    else:
        raise ValueError(
            "natten.set_memory_usage_preference allows only one of three settings: "
            "`default`, `strict`, and `unrestricted`."
        )


def get_memory_usage_preference() -> MemoryUsagePreference:
    return NattenContext.training_memory_preference


def is_memory_usage_default() -> bool:
    """Returns whether memory usage preference for KV parallelism in `"cutlass-fna"` and
    `"cutlass-fmha"` backends is the default setting.
    """
    return get_memory_usage_preference() == MemoryUsagePreference.Default


def is_memory_usage_strict() -> bool:
    """Returns whether memory usage preference for KV parallelism in `"cutlass-fna"` and
    `"cutlass-fmha"` backends is the *restricted* setting.
    """
    return get_memory_usage_preference() == MemoryUsagePreference.Strict


def is_memory_usage_unrestricted() -> bool:
    """Returns whether memory usage preference for KV parallelism in `"cutlass-fna"` and
    `"cutlass-fmha"` backends is the *unrestricted* setting.
    """
    return get_memory_usage_preference() == MemoryUsagePreference.Unrestricted


def use_deterministic_algorithms(mode: bool = True):
    NattenContext.is_deterministic_mode_enabled = mode
    if mode:
        logger.warning(
            "You're enabling NATTEN's deterministic mode. This mode does not "
            "support auto-tuning, or training with positional biases. "
            "For more information please refer to https://github.com/SHI-Labs/NATTEN/tree/main/docs"
        )


def are_deterministic_algorithms_enabled() -> bool:
    return NattenContext.is_deterministic_mode_enabled


def use_kv_parallelism_in_fused_na(mode: bool = True):
    """Sets guards for using KV Parallelism in backpropagation in `"cutlass-fna"`/`"cutlass-fmha"`
    backends.

    Warning:
        Disabling KV parallelism can significantly slow down training, particularly in
        small-batch/head and large-token problems.

    Args:
        mode: If `True`, allows KV parallelism (default setting), and otherwise disables it.
    """
    if not mode:
        NattenContext.is_kv_parallelism_enabled = False
        return

    if torch.are_deterministic_algorithms_enabled():
        logger.warning(
            "Attempted to enable KV parallelism in FNA, which is non-deterministic, "
            "but PyTorch's deterministic flag has been enabled. Ignoring..."
        )
        return

    if are_deterministic_algorithms_enabled():
        raise RuntimeError(
            "You enabled NATTEN's deterministic mode, but attempted to "
            "enable KV parallelism, which results in non-determinism. "
        )

    NattenContext.is_kv_parallelism_enabled = True


def is_kv_parallelism_in_fused_na_enabled() -> bool:
    """Returns whether KV parallelism in `"cutlass-fna"` and `"cutlass-fmha"` backends is enabled."""
    return NattenContext.is_kv_parallelism_enabled


def is_flex_compile_allowed() -> bool:
    """Returns whether compilation is allowed in `"flex-fna"` and `"flex-fmha"` backends."""
    return NattenContext.flex_compile_allowed


def is_flex_compile_backprop_allowed() -> bool:
    """Returns whether compilation for backpropagation is allowed in `"flex-fna"` and `"flex-fmha"`
    backends.
    """
    return NattenContext.flex_compile_backprop_allowed


def allow_flex_compile(mode: bool = True, backprop: bool = False):
    """Sets guards for Flex Attention + `torch.compile`.

    Allows using our Flex FNA / Flex FMHA backends with `torch.compile`, meaning you can
    pass `torch_compile=True` to the `na{1,2,3}d` or `attention` operation, along with
    `backend="flex-fna"`/`backend="flex-fmha"`, and NATTEN will compile the block-sparse mask, as
    well as the attention operation using `torch.compile` for you.

    Warning:
        We have been *unable to verify the correctness* of this setting under all of our use
        cases. We are working on raising this issue with PyTorch directly, but until then we strongly
        recommend exercising caution when using this feature.

    Danger: backprop=True is strongly discouraged!
        Allowing `torch.compile` for backpropagation (detected by checking
        `tensor.requires_grad`) is guarded separately. We strongly recommend NOT using this setting, as
        it can impact your training results.

    Args:
        mode: If `True`, enable compilation for forward pass, otherwise disable.
        backprop: If `True`, assuming compilation for forward pass is allowed, enable compilation
            for backward pass, otherwise disable.
    """
    if not mode:
        NattenContext.flex_compile_allowed = False
        NattenContext.flex_compile_backprop_allowed = False

    if not NattenContext.flex_compile_allowed:
        logger.warning(
            "You are enabling Flex Attention compilation in NATTEN. "
            "NATTEN does not allow this by default, because we cannot verify Flex's correctness in all "
            "scenarios through NATTEN's tests. By choosing to override this, you acknowledge that your "
            "results may be affected significantly. If this was not intended, please call "
            "natten.disable_flex_compile()"
            ""
        )

    NattenContext.flex_compile_allowed = True

    if backprop:
        if not NattenContext.flex_compile_backprop_allowed:
            logger.warning(
                "You are enabling using compiled Flex Attention to backpropagate. "
                "NATTEN does not allow this by default, because we cannot verify Flex's correctness in all "
                "scenarios through NATTEN's tests, and it is HIGHLY discouraged. By choosing to override "
                "this, you acknowledge that your results may be heavily impacted significantly. "
                "If this was not intended, please call "
                "natten.disable_flex_compile_backprop()"
                ""
            )
        NattenContext.flex_compile_backprop_allowed = True


def allow_flex_compile_backprop(mode: bool = True):
    """Sets guards for Flex Attention + `torch.compile` for backpropagation only.

    Args:
        mode: If `True`, enable compilation for backprop (assuming forward compilation is already
            enabled), otherwise disable.
    """
    return allow_flex_compile(is_flex_compile_allowed(), mode)


def disable_flex_compile():
    """Disallow Flex Attention + `torch.compile` entirely."""
    return allow_flex_compile(False)


def disable_flex_compile_backprop():
    """Disallow Flex Attention + `torch.compile` for backpropagation entirely."""
    return allow_flex_compile(is_flex_compile_allowed(), False)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from .._libnatten import (
    blackwell_na1d_backward,
    blackwell_na1d_forward,
    blackwell_na2d_backward,
    blackwell_na2d_forward,
    blackwell_na3d_backward,
    blackwell_na3d_forward,
)
from ..token_permute import maybe_pad, maybe_unpad, token_permute, token_unpermute
from ..types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgType,
    CausalArgTypeOrDed,
    CutlassBlackwellFnaBackwardConfigType,
    CutlassBlackwellFnaForwardConfigType,
    Dimension1DType,
    Dimension1DTypeOrDed,
    Dimension2DType,
    Dimension2DTypeOrDed,
    Dimension3DType,
    Dimension3DTypeOrDed,
    DimensionType,
    DimensionTypeOrDed,
    NoneType,
)
from ..utils.checks import check_all_args, check_args_against_input, na_tensor_checks

from .configs.checks import can_run_cutlass_blackwell_fna
from .configs.cutlass_blackwell import (
    check_cutlass_blackwell_fna_backward_config,
    check_cutlass_blackwell_fna_forward_config,
)


def make_cutlass_blackwell_fna_autograd_fn(na_dim):
    assert na_dim in [1, 2, 3]

    FORWARD_OPS = {
        1: blackwell_na1d_forward,
        2: blackwell_na2d_forward,
        3: blackwell_na3d_forward,
    }

    BACKWARD_OPS = {
        1: blackwell_na1d_backward,
        2: blackwell_na2d_backward,
        3: blackwell_na3d_backward,
    }

    class CutlassBlackwellFnaGenericAutogradFn(Function):
        @staticmethod
        @amp_fwd
        def forward(
            ctx,
            query: Tensor,
            key: Tensor,
            value: Tensor,
            kernel_size: DimensionType,
            stride: DimensionType,
            dilation: DimensionType,
            is_causal: CausalArgType,
            scale: float,
            forward_config: CutlassBlackwellFnaForwardConfigType,
            backward_config: CutlassBlackwellFnaBackwardConfigType,
            run_persistent_kernel: bool,
        ) -> Tuple[Tensor, Tensor]:
            kernel_size, stride, dilation, is_causal = check_all_args(
                na_dim, kernel_size, stride, dilation, is_causal
            )

            q_tile_shape, kv_tile_shape = forward_config

            # Token permute begin
            # Shape before padding and token permute
            qkv_shape = query.shape[1 : 1 + na_dim]

            query_pad, padding = maybe_pad(query, q_tile_shape, dilation=dilation)
            key_pad, _ = maybe_pad(key, kv_tile_shape, dilation=dilation)
            value_pad, _ = maybe_pad(value, kv_tile_shape, dilation=dilation)

            query_perm, q_shape, qR = token_permute(
                query_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            key_perm, k_shape, kR = token_permute(
                key_pad, kv_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            value_perm, v_shape, vR = token_permute(
                value_pad, kv_tile_shape, dilation=dilation, flip_tiled_dims=True
            )

            assert k_shape == v_shape
            kv_shape = k_shape
            # Token permute end

            query_perm = query_perm.contiguous()
            key_perm = key_perm.contiguous()
            value_perm = value_perm.contiguous()
            output_perm = torch.empty_like(query_perm)

            logsumexp_perm = torch.empty(
                query_perm.shape[:-1], dtype=torch.float32, device=query_perm.device
            )

            FORWARD_OPS[na_dim](
                output_perm,
                query_perm,
                key_perm,
                value_perm,
                logsumexp_perm,
                kernel_size,
                stride,
                dilation,
                is_causal,
                scale,
                q_shape,
                kv_shape,
                qkv_shape,
                0,  # num_extra_kv: disallowed for simplicity
                q_tile_shape,
                kv_tile_shape,
                run_persistent_kernel,
            )

            # Token un-permute begin
            output = maybe_unpad(
                token_unpermute(
                    output_perm,
                    q_tile_shape,
                    q_shape,
                    qR,
                    dilation=dilation,
                    flip_tiled_dims=True,
                ),
                padding,
            )
            logsumexp = maybe_unpad(
                token_unpermute(
                    logsumexp_perm.unsqueeze(-1),
                    q_tile_shape,
                    q_shape,
                    qR,
                    dilation=dilation,
                    flip_tiled_dims=True,
                ),
                padding,
            ).squeeze(-1)
            # Token un-permute end

            ctx.save_for_backward(query, key, value, logsumexp, output)
            ctx.kernel_size = kernel_size
            ctx.stride = stride
            ctx.dilation = dilation
            ctx.is_causal = is_causal
            ctx.scale = scale
            ctx.backward_config = backward_config

            return output, logsumexp

        @staticmethod
        @amp_bwd
        def backward(ctx, d_output: Tensor, d_lse: Tensor) -> Tuple[
            Tensor,
            Tensor,
            Tensor,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
            NoneType,
        ]:
            query, key, value, logsumexp, output = ctx.saved_tensors
            kernel_size, stride, dilation, is_causal, scale = (
                ctx.kernel_size,
                ctx.stride,
                ctx.dilation,
                ctx.is_causal,
                ctx.scale,
            )

            q_tile_shape, kv_tile_shape = ctx.backward_config

            # Token permute begin
            # Shape before padding and token permute
            qkv_shape = query.shape[1 : 1 + na_dim]

            query_pad, padding_q = maybe_pad(query, q_tile_shape, dilation=dilation)
            key_pad, padding_kv = maybe_pad(key, kv_tile_shape, dilation=dilation)
            value_pad, _ = maybe_pad(value, kv_tile_shape, dilation=dilation)
            logsumexp_pad, _ = maybe_pad(
                logsumexp.unsqueeze(-1), q_tile_shape, dilation=dilation
            )
            output_pad, _ = maybe_pad(output, q_tile_shape, dilation=dilation)
            d_output_pad, _ = maybe_pad(d_output, q_tile_shape, dilation=dilation)

            query_perm, q_shape, qR = token_permute(
                query_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            output_perm, o_shape, oR = token_permute(
                output_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            d_output_perm, d_o_shape, doR = token_permute(
                d_output_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            logsumexp_perm, _, _ = token_permute(
                logsumexp_pad, q_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            key_perm, k_shape, kR = token_permute(
                key_pad, kv_tile_shape, dilation=dilation, flip_tiled_dims=True
            )
            value_perm, v_shape, vR = token_permute(
                value_pad, kv_tile_shape, dilation=dilation, flip_tiled_dims=True
            )

            assert q_shape == o_shape == d_o_shape
            assert k_shape == v_shape
            kv_shape = k_shape
            # Token permute end

            query_perm = query_perm.contiguous()
            key_perm = key_perm.contiguous()
            value_perm = value_perm.contiguous()
            output_perm = output_perm.contiguous()
            d_output_perm = d_output_perm.contiguous()
            d_query_perm = torch.empty_like(query_perm)
            d_key_perm = torch.empty_like(key_perm)
            d_value_perm = torch.empty_like(value_perm)
            logsumexp_perm = logsumexp_perm.squeeze(-1)

            BACKWARD_OPS[na_dim](
                d_query_perm,
                d_key_perm,
                d_value_perm,
                query_perm,
                key_perm,
                value_perm,
                output_perm,
                d_output_perm,
                logsumexp_perm,
                kernel_size,
                stride,
                dilation,
                is_causal,
                scale,
                q_shape,
                kv_shape,
                qkv_shape,
                q_tile_shape,
                kv_tile_shape,
            )

            # Token un-permute begin
            d_query = maybe_unpad(
                token_unpermute(
                    d_query_perm,
                    q_tile_shape,
                    q_shape,
                    qR,
                    dilation=dilation,
                    flip_tiled_dims=True,
                ),
                padding_q,
            )
            d_key = maybe_unpad(
                token_unpermute(
                    d_key_perm,
                    kv_tile_shape,
                    kv_shape,
                    kR,
                    dilation=dilation,
                    flip_tiled_dims=True,
                ),
                padding_kv,
            )
            d_value = maybe_unpad(
                token_unpermute(
                    d_value_perm,
                    kv_tile_shape,
                    kv_shape,
                    vR,
                    dilation=dilation,
                    flip_tiled_dims=True,
                ),
                padding_kv,
            )
            # Token un-permute end

            assert d_query.shape == query.shape
            assert d_key.shape == key.shape
            assert d_value.shape == value.shape

            return (
                d_query,
                d_key,
                d_value,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
            )

    return CutlassBlackwellFnaGenericAutogradFn


CutlassBlackwellFna1DAutogradFn = make_cutlass_blackwell_fna_autograd_fn(1)
CutlassBlackwellFna2DAutogradFn = make_cutlass_blackwell_fna_autograd_fn(2)
CutlassBlackwellFna3DAutogradFn = make_cutlass_blackwell_fna_autograd_fn(3)


CutlassBlackwellFNAAutogradFns = {
    1: CutlassBlackwellFna1DAutogradFn,
    2: CutlassBlackwellFna2DAutogradFn,
    3: CutlassBlackwellFna3DAutogradFn,
}


def cutlass_blackwell_fna_generic(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: DimensionTypeOrDed,
    stride: DimensionTypeOrDed = 1,
    dilation: DimensionTypeOrDed = 1,
    is_causal: Optional[CausalArgTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    run_persistent_kernel: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    na_tensor_checks(query, key, value, must_match_head_dims=True)

    assert can_run_cutlass_blackwell_fna(query, key, value, raise_error=True)

    na_dim = query.dim() - 3  # batch, heads, head_dim

    kernel_size, stride, dilation, is_causal = check_all_args(
        na_dim, kernel_size, stride, dilation, is_causal
    )

    check_args_against_input(
        query,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
    )

    forward_config = check_cutlass_blackwell_fna_forward_config(
        input_tensor=query, q_tile_shape=q_tile_shape, kv_tile_shape=kv_tile_shape
    )

    backward_config = check_cutlass_blackwell_fna_backward_config(
        input_tensor=query,
        q_tile_shape=backward_q_tile_shape,
        kv_tile_shape=backward_kv_tile_shape,
    )

    scale = scale or query.shape[-1] ** -0.5

    output, lse = CutlassBlackwellFNAAutogradFns[na_dim].apply(
        query,
        key,
        value,
        kernel_size,
        stride,
        dilation,
        is_causal,
        scale,
        forward_config,
        backward_config,
        run_persistent_kernel,
    )

    if return_lse:
        return output, lse

    return output


def na1d_cutlass_blackwell_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension1DTypeOrDed,
    stride: Dimension1DTypeOrDed = 1,
    dilation: Dimension1DTypeOrDed = 1,
    is_causal: Optional[CausalArg1DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension1DType] = None,
    kv_tile_shape: Optional[Dimension1DType] = None,
    backward_q_tile_shape: Optional[Dimension1DType] = None,
    backward_kv_tile_shape: Optional[Dimension1DType] = None,
    run_persistent_kernel: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_blackwell_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        run_persistent_kernel=run_persistent_kernel,
        return_lse=return_lse,
    )


def na2d_cutlass_blackwell_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension2DTypeOrDed,
    stride: Dimension2DTypeOrDed = 1,
    dilation: Dimension2DTypeOrDed = 1,
    is_causal: Optional[CausalArg2DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension2DType] = None,
    kv_tile_shape: Optional[Dimension2DType] = None,
    backward_q_tile_shape: Optional[Dimension2DType] = None,
    backward_kv_tile_shape: Optional[Dimension2DType] = None,
    run_persistent_kernel: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_blackwell_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        run_persistent_kernel=run_persistent_kernel,
        return_lse=return_lse,
    )


def na3d_cutlass_blackwell_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension3DTypeOrDed,
    stride: Dimension3DTypeOrDed = 1,
    dilation: Dimension3DTypeOrDed = 1,
    is_causal: Optional[CausalArg3DTypeOrDed] = False,
    scale: Optional[float] = None,
    q_tile_shape: Optional[Dimension3DType] = None,
    kv_tile_shape: Optional[Dimension3DType] = None,
    backward_q_tile_shape: Optional[Dimension3DType] = None,
    backward_kv_tile_shape: Optional[Dimension3DType] = None,
    run_persistent_kernel: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    return cutlass_blackwell_fna_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        run_persistent_kernel=run_persistent_kernel,
        return_lse=return_lse,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import List, Optional

import torch  # noqa: F401
from torch import Tensor

from ....types import (
    CutlassHopperFmhaBackwardConfigType,
    CutlassHopperFmhaForwardConfigType,
    CutlassHopperFnaBackwardConfigType,
    CutlassHopperFnaForwardConfigType,
    DimensionType,
    KernelSchedule,
)
from ....utils.checks import check_tile_shape
from ....utils.device import get_device_cc


DTYPE_TO_BITS = {
    torch.float16: 16,
    torch.bfloat16: 16,
}

# TODO: notes

HOPPER_FORWARD_CONFIGS = {
    1: {
        16: {
            32: [
                (((64,), (128,)), KernelSchedule.NonPersistent),
            ],
            64: [
                (((64,), (128,)), KernelSchedule.NonPersistent),
            ],
            128: [
                (((128,), (128,)), KernelSchedule.WarpSpecializedCooperative),
                (((128,), (128,)), KernelSchedule.WarpSpecializedPingpong),
            ],
            256: [
                (((128,), (64,)), KernelSchedule.WarpSpecializedCooperative),
            ],
        },
    },
    2: {
        16: {
            32: [
                (((8, 8), (16, 8)), KernelSchedule.NonPersistent),
                (((8, 8), (8, 16)), KernelSchedule.NonPersistent),
            ],
            64: [
                (((8, 8), (16, 8)), KernelSchedule.NonPersistent),
                (((8, 8), (8, 16)), KernelSchedule.NonPersistent),
            ],
            128: [
                (((16, 8), (16, 8)), KernelSchedule.WarpSpecializedCooperative),
                (((16, 8), (16, 8)), KernelSchedule.WarpSpecializedPingpong),
            ],
            256: [
                (((16, 8), (8, 8)), KernelSchedule.WarpSpecializedCooperative),
                (((8, 16), (8, 8)), KernelSchedule.WarpSpecializedCooperative),
            ],
        },
    },
    3: {
        16: {
            32: [
                (((4, 4, 4), (4, 4, 8)), KernelSchedule.NonPersistent),
                (((4, 4, 4), (2, 8, 8)), KernelSchedule.NonPersistent),
            ],
            64: [
                (((4, 4, 4), (4, 4, 8)), KernelSchedule.NonPersistent),
                (((4, 4, 4), (2, 8, 8)), KernelSchedule.NonPersistent),
            ],
            128: [
                (((4, 4, 8), (4, 4, 8)), KernelSchedule.WarpSpecializedCooperative),
                (((4, 4, 8), (4, 4, 8)), KernelSchedule.WarpSpecializedPingpong),
                (((2, 8, 8), (2, 8, 8)), KernelSchedule.WarpSpecializedCooperative),
                (((2, 8, 8), (2, 8, 8)), KernelSchedule.WarpSpecializedPingpong),
            ],
            256: [
                (((4, 4, 8), (4, 4, 4)), KernelSchedule.WarpSpecializedCooperative),
                (((2, 8, 8), (4, 4, 4)), KernelSchedule.WarpSpecializedCooperative),
            ],
        },
    },
}

HOPPER_BACKWARD_CONFIGS = {
    1: {
        16: {
            32: [
                ((64,), (128,)),
                ((128,), (128,)),
            ],
            64: [
                ((64,), (128,)),
                ((128,), (128,)),
            ],
            128: [
                ((64,), (128,)),
            ],
        },
    },
    2: {
        16: {
            32: [
                ((8, 8), (16, 8)),
                ((8, 8), (8, 16)),
                ((16, 8), (16, 8)),
                ((16, 8), (8, 16)),
            ],
            64: [
                ((8, 8), (16, 8)),
                ((8, 8), (8, 16)),
                ((16, 8), (16, 8)),
                ((16, 8), (8, 16)),
            ],
            128: [
                ((8, 8), (16, 8)),
                ((8, 8), (8, 16)),
            ],
        },
    },
    3: {
        16: {
            32: [
                ((4, 4, 4), (4, 4, 8)),
                ((4, 4, 4), (2, 8, 8)),
                ((4, 4, 8), (4, 4, 8)),
                ((4, 4, 8), (2, 8, 8)),
            ],
            64: [
                ((4, 4, 4), (4, 4, 8)),
                ((4, 4, 4), (2, 8, 8)),
                ((4, 4, 8), (4, 4, 8)),
                ((4, 4, 8), (2, 8, 8)),
            ],
            128: [
                ((4, 4, 4), (4, 4, 8)),
                ((4, 4, 4), (2, 8, 8)),
                ((2, 4, 8), (2, 8, 8)),
                ((1, 8, 8), (2, 8, 8)),
            ],
        },
    },
}


def get_all_forward_configs(
    input_tensor: Tensor,
) -> List[CutlassHopperFnaForwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    dtype = input_tensor.dtype
    dtype_bits = DTYPE_TO_BITS[dtype]

    head_dim = input_tensor.shape[-1]

    # if dtype not in [torch.float16, torch.bfloat16]:
    if dtype_bits not in HOPPER_FORWARD_CONFIGS[na_dim]:  # type: ignore
        return []

    # if head_dim not in [32, 64, 128, 256]:
    if head_dim not in HOPPER_FORWARD_CONFIGS[na_dim][dtype_bits]:  # type: ignore
        return []

    device_cc = get_device_cc(input_tensor.device)
    if device_cc != 90:
        return []

    return HOPPER_FORWARD_CONFIGS[na_dim][dtype_bits][head_dim]  # type: ignore


def get_all_backward_configs(
    input_tensor: Tensor,
) -> List[CutlassHopperFnaBackwardConfigType]:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    dtype = input_tensor.dtype
    dtype_bits = DTYPE_TO_BITS[dtype]

    head_dim = input_tensor.shape[-1]

    # if dtype not in [torch.float16, torch.bfloat16]:
    if dtype_bits not in HOPPER_BACKWARD_CONFIGS[na_dim]:  # type: ignore
        return []

    # if head_dim not in [32, 64, 128, 256]:
    if head_dim not in HOPPER_BACKWARD_CONFIGS[na_dim][dtype_bits]:  # type: ignore
        return []

    device_cc = get_device_cc(input_tensor.device)
    if device_cc != 90:
        return []

    return HOPPER_BACKWARD_CONFIGS[na_dim][dtype_bits][head_dim]  # type: ignore


# For FMHA
def get_all_fmha_forward_configs(
    input_tensor: Tensor,
) -> List[CutlassHopperFmhaForwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    configs_multi_dim = get_all_forward_configs(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for (q_t, kv_t), _ in configs_multi_dim)

    configs_fmha = [
        ((q_t[0], kv_t[0]), sched) for (q_t, kv_t), sched in configs_multi_dim
    ]

    return configs_fmha


def get_all_fmha_backward_configs(
    input_tensor: Tensor,
) -> List[CutlassHopperFmhaBackwardConfigType]:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    configs_multi_dim = get_all_backward_configs(input_tensor)
    assert all(len(q_t) == len(kv_t) == 1 for q_t, kv_t in configs_multi_dim)

    configs_fmha = [(q_t[0], kv_t[0]) for q_t, kv_t in configs_multi_dim]

    return configs_fmha


def get_default_forward_config(
    input_tensor: Tensor,
) -> CutlassHopperFnaForwardConfigType:
    all_configs = get_all_forward_configs(input_tensor)

    if len(all_configs) < 1:
        device_cc = get_device_cc(input_tensor.device)
        raise ValueError(
            "No configs exist for this use case; Hopper FMHA/FNA does not support it: "
            f"{input_tensor.shape=}, {input_tensor.dtype=}, {device_cc=}."
        )

    return all_configs[0]


def get_default_backward_config(
    input_tensor: Tensor,
) -> CutlassHopperFnaBackwardConfigType:
    all_configs = get_all_backward_configs(input_tensor)

    if len(all_configs) < 1:
        device_cc = get_device_cc(input_tensor.device)
        raise ValueError(
            "No configs exist for this use case; Hopper FMHA/FNA does not support it: "
            f"{input_tensor.shape=}, {input_tensor.dtype=}, {device_cc=}."
        )

    return all_configs[0]


def get_default_fmha_forward_config(
    input_tensor: Tensor,
) -> CutlassHopperFmhaForwardConfigType:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    (q_t, kv_t), sched = get_default_forward_config(input_tensor)
    assert len(q_t) == len(kv_t) == 1

    return (q_t[0], kv_t[0]), sched


def get_default_fmha_backward_config(
    input_tensor: Tensor,
) -> CutlassHopperFmhaBackwardConfigType:
    if input_tensor.dim() != 4:
        raise ValueError("Only 4-D tensors are supported in FMHA.")

    q_t, kv_t = get_default_backward_config(input_tensor)
    assert len(q_t) == len(kv_t) == 1

    return q_t[0], kv_t[0]


def check_cutlass_hopper_fna_forward_config(
    input_tensor: Tensor,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    kernel_schedule: Optional[KernelSchedule] = None,
) -> CutlassHopperFnaForwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    (default_q_tile_shape, default_kv_tile_shape), default_sched = (
        get_default_forward_config(input_tensor=input_tensor)
    )
    if q_tile_shape is None and kv_tile_shape is None and kernel_schedule is None:
        return (default_q_tile_shape, default_kv_tile_shape), default_sched  # type: ignore[return-value]

    elif q_tile_shape is None and kv_tile_shape is None:
        q_tile_shape = default_q_tile_shape
        kv_tile_shape = default_kv_tile_shape

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    configs = get_all_forward_configs(input_tensor=input_tensor)

    for (q_t, kv_t), sched in configs:
        if (
            q_t == q_tile_shape
            and kv_t == kv_tile_shape
            and (kernel_schedule is None or sched == kernel_schedule)
        ):
            return (q_t, kv_t), sched  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, ((q_t, kv_t), sched) in enumerate(configs):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}, schedule={sched}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Hopper FNA-{na_dim}D. "
        f"Q tile shape {q_tile_shape}, KV tile shape {kv_tile_shape}, and schedule {kernel_schedule} "
        f"are not among the {len(configs)} configurations implementable "
        f"with CUTLASS Hopper FNA. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_hopper_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_hopper_fna_backward_config(
    input_tensor: Tensor,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
) -> CutlassHopperFnaBackwardConfigType:
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3  # batch, heads, head_dim

    if (q_tile_shape is None) ^ (kv_tile_shape is None):
        raise ValueError(
            "Please specify both q_tile_shape and kv_tile_shape, or neither one. "
            f"Got {q_tile_shape=}, {kv_tile_shape=}."
        )

    default_q_tile_shape, default_kv_tile_shape = get_default_backward_config(
        input_tensor=input_tensor
    )
    if q_tile_shape is None and kv_tile_shape is None:
        return default_q_tile_shape, default_kv_tile_shape  # type: ignore[return-value]

    elif q_tile_shape is None and kv_tile_shape is None:
        q_tile_shape = default_q_tile_shape
        kv_tile_shape = default_kv_tile_shape

    q_tile_shape = check_tile_shape(q_tile_shape)
    kv_tile_shape = check_tile_shape(kv_tile_shape)

    configs = get_all_backward_configs(input_tensor=input_tensor)

    for q_t, kv_t in configs:
        if q_t == q_tile_shape and kv_t == kv_tile_shape:
            return q_t, kv_t  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(configs):
        examples += f"\n  q_tile_shape={q_t}, kv_tile_shape={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Hopper FNA-{na_dim}D Backward. "
        f"Q tile shape {q_tile_shape} and KV tile shape {kv_tile_shape} "
        f"are not among the {len(configs)} configurations implementable "
        f"with CUTLASS Hopper FNA Backward. "
        "Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_hopper_fna(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_hopper_fmha_forward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    kernel_schedule: Optional[KernelSchedule] = None,
) -> CutlassHopperFmhaForwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    (default_q_tile_size, default_kv_tile_size), default_sched = (
        get_default_fmha_forward_config(input_tensor=input_tensor)
    )
    if q_tile_size is None and kv_tile_size is None and kernel_schedule is None:
        return (default_q_tile_size, default_kv_tile_size), default_sched

    elif q_tile_size is None and kv_tile_size is None:
        q_tile_size = default_q_tile_size
        kv_tile_size = default_kv_tile_size

    configs = get_all_fmha_forward_configs(input_tensor=input_tensor)

    for (q_t, kv_t), sched in configs:
        if (
            q_t == q_tile_size
            and kv_t == kv_tile_size
            and (kernel_schedule is None or sched == kernel_schedule)
        ):
            return (q_t, kv_t), sched  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, ((q_t, kv_t), sched) in enumerate(configs):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}, schedule={sched}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Hopper FMHA. "
        f"Q tile size {q_tile_size}, KV tile size {kv_tile_size}, and schedule {kernel_schedule} "
        f"are not among the {len(configs)} configurations implementable "
        f"with CUTLASS Hopper FMHA. "
        "Try selecting a combination from: \n"
        "  natten.get_configs_for_hopper_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )


def check_cutlass_hopper_fmha_backward_config(
    input_tensor: Tensor,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
) -> CutlassHopperFmhaBackwardConfigType:
    assert input_tensor.dim() == 4

    if (q_tile_size is None) ^ (kv_tile_size is None):
        raise ValueError(
            "Please specify both q_tile_size and kv_tile_size, or neither one. "
            f"Got {q_tile_size=}, {kv_tile_size=}."
        )

    default_q_tile_size, default_kv_tile_size = get_default_fmha_backward_config(
        input_tensor=input_tensor
    )
    if q_tile_size is None and kv_tile_size is None:
        return default_q_tile_size, default_kv_tile_size

    elif q_tile_size is None and kv_tile_size is None:
        q_tile_size = default_q_tile_size
        kv_tile_size = default_kv_tile_size

    configs = get_all_fmha_backward_configs(input_tensor=input_tensor)

    for q_t, kv_t in configs:
        if q_t == q_tile_size and kv_t == kv_tile_size:
            return q_t, kv_t  # type: ignore

    # Fail and make suggestions
    MAX_EXAMPLES = 3
    examples = ""
    for i, (q_t, kv_t) in enumerate(configs):
        examples += f"\n  q_tile_size={q_t}, kv_tile_size={kv_t}"
        if i > MAX_EXAMPLES:
            break

    raise ValueError(
        f"Invalid configuration for CUTLASS Hopper FMHA Backward. "
        f"Q tile size {q_tile_size} and KV tile size {kv_tile_size} "
        f"are not among the {len(configs)} configurations implementable "
        f"with CUTLASS Hopper FMHA Backward. "
        "Try selecting a combination from: \n"
        "  natten.get_bwd_configs_for_hopper_fmha(q, k, v)"
        "\n"
        "Here's a few examples of available combinations for your use case:\n"
        f"{examples}"
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from .._libnatten import fmha_backward, fmha_forward
from ..types import (
    CutlassFmhaBackwardConfigType,
    CutlassFmhaForwardConfigType,
    NoneType,
)
from ..utils import log
from ..utils.checks import fmha_tensor_checks

from .configs.checks import can_run_cutlass_fmha
from .configs.cutlass import (
    check_cutlass_fmha_backward_config,
    check_cutlass_fmha_forward_config,
)

logger = log.get_logger(__name__)


class CutlassFmhaAutogradFn(Function):
    @staticmethod
    @amp_fwd
    def forward(
        ctx,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        scale: float,
        forward_config: CutlassFmhaForwardConfigType,
        backward_config: CutlassFmhaBackwardConfigType,
    ) -> Tuple[Tensor, Tensor]:

        query = query.contiguous()
        key = key.contiguous()
        value = value.contiguous()

        assert query.dim() == value.dim() == 4
        assert query.shape[0] == value.shape[0]
        assert query.shape[2] == value.shape[2]

        output_shape = [query.shape[0], query.shape[1], query.shape[2], value.shape[3]]
        output = torch.empty(output_shape, device=query.device, dtype=query.dtype)

        # TODO: logsumexp should be conditional
        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        q_tile_size, kv_tile_size = forward_config
        fmha_forward(
            output,
            query,
            key,
            value,
            logsumexp,
            scale,
            q_tile_size,
            kv_tile_size,
        )

        ctx.save_for_backward(query, key, value, logsumexp, output)
        ctx.scale = scale
        ctx.backward_config = backward_config

        return output, logsumexp

    @staticmethod
    @amp_bwd
    def backward(ctx, grad_out: Tensor, grad_lse: Tensor) -> Tuple[
        Tensor,
        Tensor,
        Tensor,
        NoneType,
        NoneType,
        NoneType,
    ]:
        query, key, value, logsumexp, output = ctx.saved_tensors
        d_output = grad_out.contiguous()
        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        q_tile_size, k_tile_size, kv_splits, compute_delta_with_pt = ctx.backward_config

        num_kv_splits = kv_splits
        if kv_splits > 1 and torch.are_deterministic_algorithms_enabled():
            num_kv_splits = 1
            logger.warning(
                "You enabled PyTorch's deterministic mode, but tried to train with FNA's KV "
                "parallelism, which is non-deterministic. "
                f"Overriding {kv_splits=} to {num_kv_splits}."
            )

        fmha_backward(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            ctx.scale,
            q_tile_size,
            k_tile_size,
            num_kv_splits,
            compute_delta_with_pt,
        )

        return d_query, d_key, d_value, None, None, None


def cutlass_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: Optional[float] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
    backward_kv_splits: Optional[int] = None,
    backward_use_pt_reduction: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    fmha_tensor_checks(query, key, value, must_match_head_dims=False)

    assert can_run_cutlass_fmha(query, key, value, raise_error=True)

    forward_config = check_cutlass_fmha_forward_config(
        input_tensor=query if value.shape[-1] <= query.shape[-1] else value,
        q_tile_size=q_tile_size,
        kv_tile_size=kv_tile_size,
    )

    backward_config = check_cutlass_fmha_backward_config(
        input_tensor=key if value.shape[-1] <= key.shape[-1] else value,
        q_tile_size=backward_q_tile_size,
        kv_tile_size=backward_kv_tile_size,
        kv_splits=backward_kv_splits,
        use_pt_reduction=backward_use_pt_reduction,
    )

    scale = scale or query.shape[-1] ** -0.5

    output, lse = CutlassFmhaAutogradFn.apply(
        query,
        key,
        value,
        scale,
        forward_config,
        backward_config,
    )

    if return_lse:
        return output, lse

    return output



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from .._environment import DISABLE_TQDM


def opt_progress_bar(fn, total):
    if DISABLE_TQDM:
        return fn
    try:
        import tqdm  # type: ignore[import-untyped]

        return tqdm.tqdm(fn, total=total)

    except ImportError:
        return fn


try:
    from rich.console import Console
    from rich.table import Table

    def _print_rich_table(title, headers, values):
        table = Table(show_header=bool(title), header_style="bold", title=title)
        for h in headers:
            table.add_column(h, justify="center")
        for r in values:
            table.add_row(*r)
        return table

    def print_table(title, headers, values, has_footer=False):
        console = Console()
        console.print(_print_rich_table(title, headers, values))

except ImportError:
    # Minimal table print

    import sys

    def bold(text):
        if sys.stdout.isatty():  # Only add ANSI codes if output is a terminal
            return f"\033[1m{text}\033[0m"
        return text

    def _print_ascii_table(header, rows, title=None, footer=None):
        # Combine header, rows, footer (if any) to calculate col widths
        all_rows = [header] + rows + ([footer] if footer else [])
        columns = list(zip(*all_rows))
        col_widths = [max(len(str(cell)) for cell in col) for col in columns]

        def format_row(row, bold_row=False):
            formatted = " | ".join(
                str(cell).ljust(width) for cell, width in zip(row, col_widths)
            )
            formatted = f"| {formatted} |"
            return bold(formatted) if bold_row else formatted

        def separator(char="-", junction="+"):
            return (
                junction + junction.join(char * (w + 2) for w in col_widths) + junction
            )

        # Title
        if title:
            title_lines = title.split("\n")

            total_width = sum(col_widths) + 3 * len(col_widths) + 1
            print("|" + "=" * (total_width - 2) + "|")

            for t in title_lines:
                print("|" + str(t).center(total_width - 2) + "|")

            print("|" + "=" * (total_width - 2) + "|")

        # Top border
        print(separator("="))

        # Header
        print(format_row(header, bold_row=True))
        print(separator("="))

        # Rows
        for row in rows:
            print(format_row(row))

        # Footer (if any)
        if footer:
            print(separator("="))
            print(format_row(footer, bold_row=True))

        # Bottom border
        print(separator("="))

        print()

    def print_table(title, headers, values, has_footer=False):
        if has_footer:
            rows = values[:-1]
            footer = values[-1]
        else:
            rows = values
            footer = None

        _print_ascii_table(header=headers, rows=rows, footer=footer, title=title)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import math

import torch

from .utils import log

logger = log.get_logger(__name__)


DISABLE_PADDING_WARNING = True
TOKEN_PERMUTE_PADDING_RATIO_LIMIT_UNTIL_WARNING = 0.5


def maybe_pad(tensor, tile_shape, dilation=None):
    if tensor.dim() not in [4, 5, 6]:
        raise ValueError(
            "Expected 4D, 5D, or 6D tensor (corresponding to NA1D, 2D, 3D), "
            f"got {tensor.dim()}D input."
        )

    na_dim = tensor.dim() - 3
    assert na_dim in [1, 2, 3]

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}, " f"got {tile_shape=}."
        )

    if dilation is not None and len(dilation) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D dilation for NA{na_dim}, " f"got {dilation=}."
        )

    attn_dims = tensor.shape[1 : na_dim + 1]
    tile_shape_ = tuple(x for x in tile_shape)
    if dilation is not None:
        # NOTE: LCM?
        # tile_shape_ = tuple(math.lcm(t, d) for t, d in zip(tile_shape, dilation))
        tile_shape_ = tuple(t * d for t, d in zip(tile_shape, dilation))

    rest = tuple((x + t - 1) // t for x, t in zip(attn_dims, tile_shape_))
    residual = tuple(r * t - x for x, t, r in zip(attn_dims, tile_shape_, rest))

    assert all(res >= 0 for res in residual)

    if not DISABLE_PADDING_WARNING and any(
        res / sz > TOKEN_PERMUTE_PADDING_RATIO_LIMIT_UNTIL_WARNING
        for res, sz in zip(residual, attn_dims)
    ):
        padded_attn_dims = tuple(x + p for x, p in zip(attn_dims, residual))
        logger.warning(
            "Potentially excessive padding detected in token permute: "
            f"input shape {attn_dims} will be padded to {padded_attn_dims} to handle "
            "token permutation, which can result in excessive memory usage, and "
            "performance implications. Consider choosing your tile shapes, input shapes "
            "(and dilation if you use it) accordingly. Refer to NATTEN docs for more info."
        )

    if any(res > 0 for res in residual):
        padding = [0, 0, 0, 0]  # head_dim_left, head_dim_right, heads_left, heads_right
        for res in reversed(residual):
            padding.append(0)  # left pad
            padding.append(res)  # right pad
        tensor_padded = torch.nn.functional.pad(tensor, padding, "constant", 0)
    else:
        tensor_padded = tensor

    return tensor_padded, residual


def token_permute(tensor, tile_shape, dilation=None, flip_tiled_dims: bool = True):
    if tensor.dim() not in [4, 5, 6]:
        raise ValueError(
            "Expected 4D, 5D, or 6D tensor (corresponding to NA1D, 2D, 3D), "
            f"got {tensor.dim()}D input."
        )

    na_dim = tensor.dim() - 3
    assert na_dim in [1, 2, 3]

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}, " f"got {tile_shape=}."
        )

    dilation = dilation or tuple(1 for _ in range(na_dim))

    if len(dilation) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D dilation for NA{na_dim}, " f"got {dilation=}."
        )

    batch, *attn_dims, heads, dim = tensor.shape

    if any(
        x % d != 0 or (x // d) % t != 0
        for x, t, d in zip(attn_dims, tile_shape, dilation)
    ):
        raise ValueError(
            "Tensor must be divisible by static tile shape and dilation, but got "
            f"{tensor.shape=}, {tile_shape=}, {dilation=}."
        )

    num_dilation_groups = math.prod(dilation)
    attn_dims_post_dilation = tuple(x // d for x, d in zip(attn_dims, dilation))
    rest = tuple(x // d // t for x, t, d in zip(attn_dims, tile_shape, dilation))
    logical_divide_dims = []
    for d, r, t in zip(dilation, rest, tile_shape):
        logical_divide_dims += [r, t, d]

    # Two permutations at once:
    #  1. logical divide to tiled divide
    #  2. (optionally) flip order of tiled modes (i.e. (X,Y,Z) -> (Z,Y,X)) for compatibility with
    #     CuTe's identity layout mapping.
    permutation_idxes_r = []
    permutation_idxes_t = []
    permutation_idxes_d = []
    for i in range(na_dim):
        if flip_tiled_dims:
            permutation_idxes_r += [(na_dim - i - 1) * 3 + 1]
            permutation_idxes_t += [(na_dim - i - 1) * 3 + 2]
            permutation_idxes_d += [(na_dim - i - 1) * 3 + 3]
        else:
            permutation_idxes_r += [i * 3 + 1]
            permutation_idxes_t += [i * 3 + 2]
            permutation_idxes_d += [i * 3 + 3]

    permutation_idxes = (
        [0]
        + permutation_idxes_r
        + permutation_idxes_t
        + permutation_idxes_d
        + [na_dim * 3 + 1, na_dim * 3 + 2]
    )

    # View, not copy
    tensor_tiled = tensor.view(batch, *logical_divide_dims, heads, dim)
    assert tensor_tiled.data_ptr() == tensor.data_ptr()

    # View, not copy
    tensor_permuted = tensor_tiled.permute(*permutation_idxes)
    assert tensor_permuted.data_ptr() == tensor_tiled.data_ptr()

    # Reshape back and copy
    tensor_flatten = tensor_permuted.reshape(
        batch, math.prod(attn_dims_post_dilation), num_dilation_groups * heads, dim
    ).contiguous()
    # NOTE: token permute without dilation is a no-op for 1-D
    # assert na_dim == 1 or tensor_flatten.data_ptr() != tensor_permuted.data_ptr()
    assert tensor_flatten.is_contiguous()

    return tensor_flatten, attn_dims_post_dilation, rest


def token_unpermute(
    tensor,
    tile_shape,
    orig_shape,
    rest_shape,
    dilation=None,
    flip_tiled_dims: bool = True,
):
    if tensor.dim() != 4:
        raise ValueError(f"Expected flattened 4D tensor, got {tensor.dim()}D input.")

    na_dim = len(orig_shape)
    assert na_dim in [1, 2, 3]

    if len(tile_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D tiler for NA{na_dim}, " f"got {tile_shape=}."
        )

    if len(rest_shape) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D rest shape for NA{na_dim}, " f"got {rest_shape=}."
        )

    dilation = dilation or tuple(1 for _ in range(na_dim))

    if len(dilation) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D dilation for NA{na_dim}, " f"got {dilation=}."
        )

    num_dilation_groups = math.prod(dilation)

    batch, seqlen, heads, dim = tensor.shape

    if heads % num_dilation_groups != 0:
        raise ValueError(
            "Expected number of heads in token-permuted tensor to be divisible by "
            f"number of dilation groups {num_dilation_groups} ({dilation=}), got {heads=}."
        )

    heads_actual = heads // num_dilation_groups

    orig_shape_pre_dilation = tuple(x * d for x, d in zip(orig_shape, dilation))

    # View, not copy
    rest_shape_ = reversed(rest_shape) if flip_tiled_dims else rest_shape
    tile_shape_ = reversed(tile_shape) if flip_tiled_dims else tile_shape
    dilation_ = reversed(dilation) if flip_tiled_dims else dilation
    tensor_tiled = tensor.view(
        batch, *rest_shape_, *tile_shape_, *dilation_, heads_actual, dim
    )
    assert tensor_tiled.data_ptr() == tensor.data_ptr()

    # Undo permutation
    permutation_idxes = [0]
    for i in range(na_dim):
        if flip_tiled_dims:
            permutation_idxes += [na_dim - i, 2 * na_dim - i, 3 * na_dim - i]
        else:
            permutation_idxes += [i + 1, na_dim + i + 1, 2 * na_dim + i + 1]

    permutation_idxes += [na_dim * 3 + 1, na_dim * 3 + 2]

    # View, not copy
    tensor_permuted = tensor_tiled.permute(*permutation_idxes)
    assert tensor_permuted.data_ptr() == tensor_tiled.data_ptr()

    # Reshape back and copy
    out = tensor_permuted.reshape(
        batch, *orig_shape_pre_dilation, heads_actual, dim
    ).contiguous()
    # NOTE: token permute without dilation is a no-op for 1-D
    # assert na_dim == 1 or out.data_ptr() != tensor_permuted.data_ptr()
    assert out.is_contiguous()

    return out


def maybe_unpad(tensor, padding):
    if tensor.dim() not in [4, 5, 6]:
        raise ValueError(
            "Expected 4D, 5D, or 6D tensor (corresponding to NA1D, 2D, 3D), "
            f"got {tensor.dim()}D input."
        )

    na_dim = tensor.dim() - 3
    assert na_dim in [1, 2, 3]

    if len(padding) != na_dim:
        raise ValueError(
            f"Expected {na_dim}D padding shape for NA{na_dim}, " f"got {padding=}."
        )

    attn_dims = tensor.shape[1 : na_dim + 1]

    # Slice
    if any(p for p in padding):
        assert all(p >= 0 for p in padding)

        orig_lens = tuple(x - p for x, p in zip(attn_dims, padding))

        # TODO: there must be a beter way
        if len(orig_lens) == 1:
            x = orig_lens[0]
            return tensor[:, :x].contiguous()
        elif len(orig_lens) == 2:
            x, y = orig_lens
            return tensor[:, :x, :y].contiguous()
        elif len(orig_lens) == 3:
            x, y, z = orig_lens
            return tensor[:, :x, :y, :z].contiguous()
        else:
            raise NotImplementedError()

    return tensor



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import os

import torch

from ._libnatten import HAS_LIBNATTEN  # noqa: F401
from .utils.device import get_device_cc

_IS_CUDA_AVAILABLE = torch.cuda.is_available()

_TORCH_VERSION = [int(x) for x in torch.__version__.split(".")[:2]]

_IS_TORCH_COMPILE_SUPPORTED = _TORCH_VERSION >= [2, 6] and get_device_cc() >= 70

# Unit tests
_RUN_EXTENDED_TESTS = bool(os.getenv("NATTEN_RUN_EXTENDED_TESTS", "0") == "1")

# Profiler
DISABLE_TQDM = bool(os.getenv("NATTEN_DISABLE_TQDM", "0") == "1")



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import time
from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor
from torch.profiler import profile as torch_profile, ProfilerActivity

from natten.functional import neighborhood_attention_generic
from natten.types import DimensionType, KernelSchedule
from natten.utils import log
from natten.utils.checks import check_all_args

from .formatting import convert_to_natten_profiler_ops, Result
from .ops import sdpa

from .problem import Problem

logger = log.get_logger(__name__)

IS_CUDA = torch.cuda.is_available()
torch_device = "cuda" if IS_CUDA else "cpu"
profiler_activity_tag = ProfilerActivity.CUDA if IS_CUDA else ProfilerActivity.CPU


def init_tensors(
    problem: Problem, flatten_sequence: bool, heads_last: bool
) -> Tuple[Tensor, Tensor, Tensor, Tensor, Optional[Tuple[Tensor, Tensor]]]:
    q, k, v, d_out = problem.make_qkvo_tensors(
        device=torch_device,
        requires_grad=False,
        heads_last=heads_last,
        flatten=flatten_sequence,
    )

    add_kv = None
    if problem.has_additional_kv:
        assert not flatten_sequence
        add_k, add_v = problem.make_additional_kv_tensors(
            device=torch_device, requires_grad=False, heads_last=heads_last
        )
        add_kv = (add_k, add_v)

    return q, k, v, d_out, add_kv


# Doesn't profile; uses cuda events and only reports a runtime
def measure_natten_runtime(
    problem: Problem,
    warmup_steps: int,
    backend: Optional[str] = None,
    fmha_backend: Optional[str] = None,
    disable_backward: bool = False,
    # Perf-related args
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    backward_kv_splits: Optional[DimensionType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
    # FMHA args
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
) -> float:

    window_size, stride, dilation, is_causal = check_all_args(
        problem.na_dim,
        problem.window_size,
        problem.stride,
        problem.dilation,
        problem.is_causal,
    )

    query, key, value, d_out, additional_kv = init_tensors(
        problem, flatten_sequence=False, heads_last=True
    )

    torch.set_grad_enabled(not disable_backward)

    def run_ops(
        query,
        key,
        value,
        d_out,
        window_size,
        stride,
        dilation,
        is_causal,
        disable_backward,
        additional_kv,
    ):
        query.requires_grad = not disable_backward
        key.requires_grad = not disable_backward
        value.requires_grad = not disable_backward
        additional_keys, additional_values = (
            (None, None) if additional_kv is None else additional_kv
        )
        if additional_kv is not None:
            additional_keys.requires_grad = not disable_backward
            additional_values.requires_grad = not disable_backward

        out = neighborhood_attention_generic(
            query,
            key,
            value,
            kernel_size=window_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            additional_keys=additional_keys,
            additional_values=additional_values,
            backend=backend,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            backward_kv_splits=backward_kv_splits,
            backward_use_pt_reduction=backward_use_pt_reduction,
            run_persistent_kernel=run_persistent_kernel,
            kernel_schedule=kernel_schedule,
            torch_compile=torch_compile,
            attention_kwargs={
                "backend": fmha_backend,
                "kernel_schedule": kernel_schedule,
                "q_tile_size": q_tile_size,
                "kv_tile_size": kv_tile_size,
                "backward_q_tile_size": backward_q_tile_size,
                "backward_kv_tile_size": backward_kv_tile_size,
            },
        )

        if not disable_backward:
            out.backward(d_out)

    for _ in range(warmup_steps):
        with torch.no_grad():
            q, k, v, do = (
                query.clone(),
                key.clone(),
                value.clone(),
                d_out.clone(),
            )
            add_kv = (
                None
                if additional_kv is None
                else (additional_kv[0].clone(), additional_kv[1].clone())
            )
        run_ops(
            q,
            k,
            v,
            do,
            window_size,
            stride,
            dilation,
            is_causal,
            disable_backward,
            add_kv,
        )

    with torch.no_grad():
        q, k, v, do = (
            query.clone(),
            key.clone(),
            value.clone(),
            d_out.clone(),
        )
        add_kv = (
            None
            if additional_kv is None
            else (additional_kv[0].clone(), additional_kv[1].clone())
        )

    if IS_CUDA:
        start_event = torch.cuda.Event(enable_timing=True)
        end_event = torch.cuda.Event(enable_timing=True)
        start_event.record()
    else:
        start_time = time.time()

    run_ops(
        q,
        k,
        v,
        do,
        window_size,
        stride,
        dilation,
        is_causal,
        disable_backward,
        add_kv,
    )

    if IS_CUDA:
        end_event.record()
        torch.cuda.synchronize()
        elapsed_time_ms = start_event.elapsed_time(end_event)
    else:
        elapsed_time_ms = (time.time() - start_time) * 1e3

    return elapsed_time_ms


def _profile_na_with_torch(
    problem: Problem,
    warmup_steps: int,
    backend: Optional[str] = None,
    fmha_backend: Optional[str] = None,
    disable_backward: bool = False,
    # Perf-related args
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    backward_kv_splits: Optional[DimensionType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
    # FMHA args
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
) -> torch_profile:

    window_size, stride, dilation, is_causal = check_all_args(
        problem.na_dim,
        problem.window_size,
        problem.stride,
        problem.dilation,
        problem.is_causal,
    )

    query, key, value, d_out, additional_kv = init_tensors(
        problem, flatten_sequence=False, heads_last=True
    )

    torch.set_grad_enabled(not disable_backward)

    def run_ops(
        query,
        key,
        value,
        d_out,
        window_size,
        stride,
        dilation,
        is_causal,
        disable_backward,
        additional_kv,
    ):
        if IS_CUDA:
            torch.cuda.synchronize()
        query.requires_grad = not disable_backward
        key.requires_grad = not disable_backward
        value.requires_grad = not disable_backward
        additional_keys, additional_values = (
            (None, None) if additional_kv is None else additional_kv
        )
        if additional_kv is not None:
            additional_keys.requires_grad = not disable_backward
            additional_values.requires_grad = not disable_backward

        out = neighborhood_attention_generic(
            query,
            key,
            value,
            kernel_size=window_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            additional_keys=additional_keys,
            additional_values=additional_values,
            backend=backend,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            backward_kv_splits=backward_kv_splits,
            backward_use_pt_reduction=backward_use_pt_reduction,
            run_persistent_kernel=run_persistent_kernel,
            kernel_schedule=kernel_schedule,
            torch_compile=torch_compile,
            attention_kwargs={
                "backend": fmha_backend,
                "kernel_schedule": kernel_schedule,
                "q_tile_size": q_tile_size,
                "kv_tile_size": kv_tile_size,
                "backward_q_tile_size": backward_q_tile_size,
                "backward_kv_tile_size": backward_kv_tile_size,
            },
        )

        if IS_CUDA:
            torch.cuda.synchronize()
        if not disable_backward:
            out.backward(d_out)
            if IS_CUDA:
                torch.cuda.synchronize()

    for _ in range(warmup_steps):
        with torch.no_grad():
            q, k, v, do = (
                query.clone(),
                key.clone(),
                value.clone(),
                d_out.clone(),
            )
            add_kv = (
                None
                if additional_kv is None
                else (additional_kv[0].clone(), additional_kv[1].clone())
            )
        run_ops(
            q,
            k,
            v,
            do,
            window_size,
            stride,
            dilation,
            is_causal,
            disable_backward,
            add_kv,
        )

    with torch.no_grad():
        q, k, v, do = (
            query.clone(),
            key.clone(),
            value.clone(),
            d_out.clone(),
        )
        add_kv = (
            None
            if additional_kv is None
            else (additional_kv[0].clone(), additional_kv[1].clone())
        )

    with torch_profile(activities=[profiler_activity_tag]) as prof:
        run_ops(
            q,
            k,
            v,
            do,
            window_size,
            stride,
            dilation,
            is_causal,
            disable_backward,
            add_kv,
        )

    return prof


def profile_na_with_torch(
    problem: Problem,
    warmup_steps: int = 10,
    backend: Optional[str] = None,
    fmha_backend: Optional[str] = None,
    disable_backward: bool = False,
    # Perf-related args
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    backward_kv_splits: Optional[DimensionType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
    # Debug
    debug_report_prof_result: bool = False,
    # FMHA args
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
) -> Optional[List[Result]]:
    profile_result = _profile_na_with_torch(
        problem,
        warmup_steps=warmup_steps,
        backend=backend,
        fmha_backend=fmha_backend,
        disable_backward=disable_backward,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        run_persistent_kernel=run_persistent_kernel,
        kernel_schedule=kernel_schedule,
        torch_compile=torch_compile,
        q_tile_size=q_tile_size,
        kv_tile_size=kv_tile_size,
        backward_q_tile_size=backward_q_tile_size,
        backward_kv_tile_size=backward_kv_tile_size,
    )

    if debug_report_prof_result:
        logger.debug(profile_result.key_averages().table())

    out = convert_to_natten_profiler_ops(profile_result)
    assert out is not None

    return out


def _profile_fmha_with_torch(
    problem: Problem,
    warmup_steps: int,
    backend: Optional[str] = "cudnn",
    disable_backward: Optional[bool] = False,
) -> torch_profile:
    torch.set_grad_enabled(not disable_backward)

    assert (
        not problem.has_additional_kv
    ), "Profiling SDPA with additional KV is not supported."

    query, key, value, d_out, additional_kv = init_tensors(
        problem, flatten_sequence=True, heads_last=False  # torch SDPA is heads first :(
    )

    def run_ops(query, key, value, d_out, backend):
        query.requires_grad = not disable_backward
        key.requires_grad = not disable_backward
        value.requires_grad = not disable_backward
        out = sdpa(query, key, value, backend=backend)
        if not disable_backward:
            out.backward(d_out)

    logger.warning(
        "Running standard self attention; ignoring NA parameters window size, stride, dilation, causal, etc."
    )

    for _ in range(warmup_steps):
        run_ops(
            query.clone(),
            key.clone(),
            value.clone(),
            d_out.clone(),
            backend=backend,
        )

    with torch_profile(activities=[profiler_activity_tag]) as prof:
        run_ops(
            query.clone(),
            key.clone(),
            value.clone(),
            d_out.clone(),
            backend=backend,
        )

    return prof


def profile_fmha_with_torch(
    problem: Problem,
    warmup_steps: int = 10,
    backend: Optional[str] = "cudnn",
    disable_backward: Optional[bool] = False,
    # Debug
    debug_report_prof_result: bool = False,
) -> Optional[List[Result]]:
    profile_result = _profile_fmha_with_torch(
        problem,
        warmup_steps=warmup_steps,
        backend=backend,
        disable_backward=disable_backward,
    )

    if debug_report_prof_result:
        logger.debug(profile_result.key_averages().table())

    out = convert_to_natten_profiler_ops(profile_result)
    assert out is not None

    return out



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from enum import Enum
from typing import Tuple, Union

NoneType = type(None)

Dimension1DType = Tuple[int]
Dimension2DType = Tuple[int, int]
Dimension3DType = Tuple[int, int, int]

CausalArg1DType = Tuple[bool]
CausalArg2DType = Tuple[bool, bool]
CausalArg3DType = Tuple[bool, bool, bool]

# NOTE: switch to | when < 3.10 support is dropped
Dimension1DTypeOrDed = Union[int, Dimension1DType]
Dimension2DTypeOrDed = Union[int, Dimension2DType]
Dimension3DTypeOrDed = Union[int, Dimension3DType]

CausalArg1DTypeOrDed = Union[bool, CausalArg1DType]
CausalArg2DTypeOrDed = Union[bool, CausalArg2DType]
CausalArg3DTypeOrDed = Union[bool, CausalArg3DType]

DimensionType = Union[Dimension1DType, Dimension2DType, Dimension3DType]
CausalArgType = Union[CausalArg1DType, CausalArg2DType, CausalArg3DType]

DimensionTypeOrDed = Union[int, DimensionType]
CausalArgTypeOrDed = Union[bool, CausalArgType]

# (query_tile_shape, kv_tile_shape)
QKTileShapeType = Union[
    Tuple[Dimension1DType, Dimension1DType],
    Tuple[Dimension2DType, Dimension2DType],
    Tuple[Dimension3DType, Dimension3DType],
]


# TODO: Only applies to Hopper FMHA/FNA for now -- extend to other applicable kernels
class KernelSchedule(Enum):
    NonPersistent = 0
    WarpSpecializedCooperative = 1
    WarpSpecializedPingpong = 2


CutlassFnaForwardConfigType = QKTileShapeType
CutlassBlackwellFnaForwardConfigType = QKTileShapeType
CutlassBlackwellFnaBackwardConfigType = QKTileShapeType
CutlassHopperFnaForwardConfigType = Tuple[QKTileShapeType, KernelSchedule]
CutlassHopperFnaBackwardConfigType = QKTileShapeType
FlexFnaForwardConfigType = QKTileShapeType

# (query_tile_shape, kv_tile_shape, num_kv_splits, use_torch_to_compute_delta)
CutlassFnaBackwardConfigType = Union[
    Tuple[Dimension1DType, Dimension1DType, Dimension1DType, bool],
    Tuple[Dimension2DType, Dimension2DType, Dimension2DType, bool],
    Tuple[Dimension3DType, Dimension3DType, Dimension3DType, bool],
]

# FMHA configs
FmhaForwardConfigType = Tuple[int, int]

CutlassFmhaForwardConfigType = FmhaForwardConfigType
CutlassFmhaBackwardConfigType = Tuple[int, int, int, bool]

FlexFmhaForwardConfigType = FmhaForwardConfigType
CutlassBlackwellFmhaForwardConfigType = FmhaForwardConfigType
CutlassBlackwellFmhaBackwardConfigType = FmhaForwardConfigType
CutlassHopperFmhaForwardConfigType = Tuple[FmhaForwardConfigType, KernelSchedule]
CutlassHopperFmhaBackwardConfigType = FmhaForwardConfigType



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


from typing import List

from ..utils import log

logger = log.get_logger(__name__)

import torch  # noqa: F401
from torch import Tensor

from .blackwell_fmha import cutlass_blackwell_fmha
from .blackwell_fna import (
    cutlass_blackwell_fna_generic,
    na1d_cutlass_blackwell_fna,
    na2d_cutlass_blackwell_fna,
    na3d_cutlass_blackwell_fna,
)

from .configs import (
    get_bwd_configs_for_cutlass_blackwell_fmha,
    get_bwd_configs_for_cutlass_blackwell_fna,
    get_bwd_configs_for_cutlass_fmha,
    get_bwd_configs_for_cutlass_fna,
    get_bwd_configs_for_cutlass_hopper_fmha,
    get_bwd_configs_for_cutlass_hopper_fna,
    get_configs_for_cutlass_blackwell_fmha,
    get_configs_for_cutlass_blackwell_fna,
    get_configs_for_cutlass_fmha,
    get_configs_for_cutlass_fna,
    get_configs_for_cutlass_hopper_fmha,
    get_configs_for_cutlass_hopper_fna,
    get_configs_for_flex_fmha,
    get_configs_for_flex_fna,
)

from .configs.checks import (
    can_run_cutlass_blackwell_fmha,
    can_run_cutlass_blackwell_fna,
    can_run_cutlass_fna,
    can_run_cutlass_hopper_fmha,
    can_run_cutlass_hopper_fna,
    can_run_flex_attention,
)

from .flex import flex_fmha, flex_fna_generic, na1d_flex, na2d_flex, na3d_flex
from .fmha import can_run_cutlass_fmha, cutlass_fmha
from .fna import (
    cutlass_fna_generic,
    na1d_cutlass_fna,
    na2d_cutlass_fna,
    na3d_cutlass_fna,
)
from .hopper_fmha import cutlass_hopper_fmha
from .hopper_fna import (
    cutlass_hopper_fna_generic,
    na1d_cutlass_hopper_fna,
    na2d_cutlass_hopper_fna,
    na3d_cutlass_hopper_fna,
)


def choose_backend(
    query: Tensor, key: Tensor, value: Tensor, torch_compile: bool
) -> str:
    if can_run_cutlass_blackwell_fna(query, key, value):
        logger.debug("Backend not set; picked Blackwell FNA kernel.")
        return "blackwell-fna"

    if can_run_cutlass_hopper_fna(query, key, value):
        logger.debug("Backend not set; picked Hopper FNA kernel.")
        return "hopper-fna"

    if can_run_cutlass_fna(query, key, value):
        logger.debug("Backend not set; picked CUTLASS (2.X) FNA kernel.")
        return "cutlass-fna"

    if can_run_flex_attention(query, key, value, torch_compile=torch_compile):
        logger.debug("Backend not set; picked Flex Attention kernel.")
        return "flex-fna"

    raise NotImplementedError(
        "NATTEN could not find a suitable backend for this use case. "
        "Run with NATTEN_LOG_LEVEL=DEBUG to find out why."
    )


def choose_fmha_backend(
    query: Tensor, key: Tensor, value: Tensor, torch_compile: bool
) -> str:
    if can_run_cutlass_blackwell_fmha(query, key, value):
        logger.debug("Backend not set; picked Blackwell FMHA kernel.")
        return "blackwell-fmha"

    if can_run_cutlass_hopper_fmha(query, key, value):
        logger.debug("Backend not set; picked Hopper FMHA kernel.")
        return "hopper-fmha"

    if can_run_cutlass_fmha(query, key, value):
        logger.debug("Backend not set; picked CUTLASS (2.X) FMHA kernel.")
        return "cutlass-fmha"

    if can_run_flex_attention(query, key, value, torch_compile=torch_compile):
        logger.debug("Backend not set; picked Flex Attention kernel.")
        return "flex-fmha"

    raise NotImplementedError(
        "NATTEN could not find a suitable backend for this FMHA use case. "
        "Run with NATTEN_LOG_LEVEL=DEBUG to find out why."
    )


def get_compatible_backends(
    query: Tensor, key: Tensor, value: Tensor, torch_compile: bool
) -> List[str]:
    compatible_backends = []
    if can_run_cutlass_blackwell_fna(query, key, value):
        compatible_backends.append("blackwell-fna")

    if can_run_cutlass_hopper_fna(query, key, value):
        compatible_backends.append("hopper-fna")

    if can_run_cutlass_fna(query, key, value):
        compatible_backends.append("cutlass-fna")

    if can_run_flex_attention(query, key, value, torch_compile=torch_compile):
        compatible_backends.append("flex-fna")

    return compatible_backends


def get_compatible_fmha_backends(
    query: Tensor, key: Tensor, value: Tensor, torch_compile: bool
) -> List[str]:
    compatible_backends = []
    if can_run_cutlass_blackwell_fmha(query, key, value):
        compatible_backends.append("blackwell-fmha")

    if can_run_cutlass_hopper_fmha(query, key, value):
        compatible_backends.append("hopper-fmha")

    if can_run_cutlass_fmha(query, key, value):
        compatible_backends.append("cutlass-fmha")

    if can_run_flex_attention(query, key, value, torch_compile=torch_compile):
        compatible_backends.append("flex-fmha")

    return compatible_backends


__all__ = [
    "can_run_cutlass_fmha",
    "can_run_cutlass_fna",
    "can_run_cutlass_blackwell_fmha",
    "can_run_cutlass_blackwell_fna",
    "can_run_cutlass_hopper_fmha",
    "can_run_cutlass_hopper_fna",
    "can_run_flex_attention",
    "cutlass_fmha",
    "cutlass_fna_generic",
    "na1d_cutlass_fna",
    "na2d_cutlass_fna",
    "na3d_cutlass_fna",
    "cutlass_blackwell_fmha",
    "cutlass_blackwell_fna_generic",
    "cutlass_hopper_fmha",
    "cutlass_hopper_fna_generic",
    "na1d_cutlass_blackwell_fna",
    "na2d_cutlass_blackwell_fna",
    "na3d_cutlass_blackwell_fna",
    "flex_fmha",
    "flex_fna_generic",
    "na1d_flex",
    "na2d_flex",
    "na3d_flex",
    "na1d_cutlass_hopper_fna",
    "na2d_cutlass_hopper_fna",
    "na3d_cutlass_hopper_fna",
    "get_bwd_configs_for_cutlass_fmha",
    "get_bwd_configs_for_cutlass_fna",
    "get_bwd_configs_for_cutlass_blackwell_fmha",
    "get_bwd_configs_for_cutlass_blackwell_fna",
    "get_configs_for_cutlass_blackwell_fmha",
    "get_configs_for_cutlass_blackwell_fna",
    "get_configs_for_cutlass_fmha",
    "get_configs_for_cutlass_fna",
    "get_configs_for_cutlass_hopper_fmha",
    "get_bwd_configs_for_cutlass_hopper_fmha",
    "get_configs_for_cutlass_hopper_fna",
    "get_bwd_configs_for_cutlass_hopper_fna",
    "get_configs_for_flex_fmha",
    "get_configs_for_flex_fna",
]
